"""
Exposure Value Computation
==========================

Defines the exposure value computation objects:

-   :func:`colour_hdri.average_luminance`
-   :func:`colour_hdri.average_illuminance`
-   :func:`colour_hdri.luminance_to_exposure_value`
-   :func:`colour_hdri.illuminance_to_exposure_value`
-   :func:`colour_hdri.adjust_exposure`

References
----------
-   :cite:`Wikipediabj` : Wikipedia. (n.d.). EV as a measure of luminance and
    illuminance. Retrieved November 14, 2015, from
    https://en.wikipedia.org/wiki/Exposure_value#\
EV_as_a_measure_of_luminance_and_illuminance
"""

from __future__ import annotations

import numpy as np

from colour.hints import FloatingOrArrayLike, FloatingOrNDArray

from colour.utilities import as_float, as_float_array

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "average_luminance",
    "average_illuminance",
    "luminance_to_exposure_value",
    "illuminance_to_exposure_value",
    "adjust_exposure",
]


def average_luminance(
    N: FloatingOrArrayLike,
    t: FloatingOrArrayLike,
    S: FloatingOrArrayLike,
    k: FloatingOrArrayLike = 12.5,
) -> FloatingOrNDArray:
    """
    Compute the average luminance :math:`L` in :math:`cd\\cdot m^{-2}` from
    given relative aperture *F-Number* :math:`N`, *Exposure Time* :math:`t`,
    *ISO* arithmetic speed :math:`S` and *reflected light calibration constant*
    :math:`k`.

    Parameters
    ----------
    N
        Relative aperture *F-Number* :math:`N`.
    t
       *Exposure Time* :math:`t`.
    S
        *ISO* arithmetic speed :math:`S`.
    k
        *Reflected light calibration constant* :math:`k`.
        *ISO 2720:1974* recommends a range for :math:`k` of 10.6 to 13.4 with
        luminance in :math:`cd\\cdot m^{-2}`. Two values for :math:`k` are in
        common use: 12.5 (Canon, Nikon, and Sekonic) and 14 (Minolta, Kenko,
        and Pentax).

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Average luminance :math:`L` in :math:`cd\\cdot m^{-2}`.

    References
    ----------
    :cite:`Wikipediabj`

    Examples
    --------
    >>> average_luminance(8, 1, 100)
    8.0
    """

    N = as_float_array(N)
    t = as_float_array(t)
    S = as_float_array(S)
    k = as_float_array(k)

    L = N**2 / t / S * k

    return as_float(L)


def average_illuminance(
    N: FloatingOrArrayLike,
    t: FloatingOrArrayLike,
    S: FloatingOrArrayLike,
    c: FloatingOrArrayLike = 250,
) -> FloatingOrNDArray:
    """
    Compute the average illuminance :math:`E` in :math:`Lux` from given
    relative aperture *F-Number* :math:`N`, *Exposure Time* :math:`t`, *ISO*
    arithmetic speed :math:`S` and *incident light calibration constant*
    :math:`c`.

    Parameters
    ----------
    N
        Relative aperture *F-Number* :math:`N`.
    t
       *Exposure Time* :math:`t`.
    S
        *ISO* arithmetic speed :math:`S`.
    c
        *Incident light calibration constant* :math:`c`.
        With a flat receptor, *ISO 2720:1974* recommends a range for
        :math:`c`. of 240 to 400 with illuminance in :math:`Lux`; a value of
        250 is commonly used. With a hemispherical receptor, *ISO 2720:1974*
        recommends a range for :math:`c` of 320 to 540 with illuminance in
        :math:`Lux`; in practice, values typically are between 320 (Minolta)
        and 340 (Sekonic).

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Average illuminance :math:`E` in :math:`Lux`.

    References
    ----------
    :cite:`Wikipediabj`

    Examples
    --------
    >>> average_illuminance(8, 1, 100)
    160.0
    """

    N = as_float_array(N)
    t = as_float_array(t)
    S = as_float_array(S)
    c = as_float_array(c)

    E = N**2 / t / S * c

    return as_float(E)


def luminance_to_exposure_value(
    L: FloatingOrArrayLike,
    S: FloatingOrArrayLike,
    k: FloatingOrArrayLike = 12.5,
) -> FloatingOrNDArray:
    """
    Compute the exposure value :math:`EV` from given scene luminance
    :math:`L` in :math:`cd\\cdot m^{-2}`, *ISO* arithmetic speed :math:`S` and
    *reflected light calibration constant* :math:`k`.

    Parameters
    ----------
    L
        Scene luminance :math:`L` in :math:`cd\\cdot m^{-2}`.
    S
        *ISO* arithmetic speed :math:`S`.
    k
        *Reflected light calibration constant* :math:`k`.
        *ISO 2720:1974* recommends a range for :math:`k` of 10.6 to 13.4 with
        luminance in :math:`cd\\cdot m^{-2}`. Two values for :math:`k` are in
        common use: 12.5 (Canon, Nikon, and Sekonic) and 14 (Minolta, Kenko,
        and Pentax).

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Exposure value :math:`EV`.

    Notes
    -----
    -   The exposure value :math:`EV` indicates a combination of camera
        settings rather than the focal plane exposure, i.e. luminous exposure,
        photometric exposure, :math:`H`. The focal plane exposure is
        time-integrated illuminance.

    References
    ----------
    :cite:`Wikipediabj`

    Examples
    --------
    >>> luminance_to_exposure_value(0.125, 100)
    0.0
    """

    L = as_float_array(L)
    S = as_float_array(S)
    k = as_float_array(k)

    EV = np.log2(L * S / k)

    return as_float(EV)


def illuminance_to_exposure_value(
    E: FloatingOrArrayLike,
    S: FloatingOrArrayLike,
    c: FloatingOrArrayLike = 250,
) -> FloatingOrNDArray:
    """
    Compute the exposure value :math:`EV` from given scene illuminance
    :math:`E` in :math:`Lux`, *ISO* arithmetic speed :math:`S` and
    *incident light calibration constant* :math:`c`.

    Parameters
    ----------
    E
        Scene illuminance :math:`E` in :math:`Lux`.
    S
        *ISO* arithmetic speed :math:`S`.
    c
        *Incident light calibration constant* :math:`c`.
        With a flat receptor, *ISO 2720:1974* recommends a range for
        :math:`c`. of 240 to 400 with illuminance in :math:`Lux`; a value of
        250 is commonly used. With a hemispherical receptor, *ISO 2720:1974*
        recommends a range for :math:`c` of 320 to 540 with illuminance in
        :math:`Lux`; in practice, values typically are between 320 (Minolta)
        and 340 (Sekonic).

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Exposure value :math:`EV`.

    Notes
    -----
    -   The exposure value :math:`EV` indicates a combination of camera
        settings rather than the focal plane exposure, i.e. luminous exposure,
        photometric exposure, :math:`H`. The focal plane exposure is
        time-integrated illuminance.

    References
    ----------
    :cite:`Wikipediabj`

    Examples
    --------
    >>> illuminance_to_exposure_value(2.5, 100)
    0.0
    """

    E = as_float_array(E)
    S = as_float_array(S)
    c = as_float_array(c)

    EV = np.log2(E * S / c)

    return as_float(EV)


def adjust_exposure(
    a: FloatingOrArrayLike, EV: FloatingOrArrayLike
) -> FloatingOrNDArray:
    """
    Adjust given array exposure using given :math:`EV` exposure value.

    Parameters
    ----------
    a
        Array to adjust the exposure.
    EV
        Exposure adjustment value.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Exposure adjusted array.

    Examples
    --------
    >>> adjust_exposure(np.array([0.25, 0.5, 0.75, 1]), 1)
    array([ 0.5,  1. ,  1.5,  2. ])
    """

    a = as_float_array(a)
    EV = as_float_array(EV)

    return as_float(a * pow(2, EV))



from .operators import (
    tonemapping_operator_simple,
    tonemapping_operator_normalisation,
    tonemapping_operator_gamma,
    tonemapping_operator_logarithmic,
    tonemapping_operator_exponential,
    tonemapping_operator_logarithmic_mapping,
    tonemapping_operator_exponentiation_mapping,
    tonemapping_operator_Schlick1994,
    tonemapping_operator_Tumblin1999,
    tonemapping_operator_Reinhard2004,
    tonemapping_operator_filmic,
)

__all__ = [
    "tonemapping_operator_simple",
    "tonemapping_operator_normalisation",
    "tonemapping_operator_gamma",
    "tonemapping_operator_logarithmic",
    "tonemapping_operator_exponential",
    "tonemapping_operator_logarithmic_mapping",
    "tonemapping_operator_exponentiation_mapping",
    "tonemapping_operator_Schlick1994",
    "tonemapping_operator_Tumblin1999",
    "tonemapping_operator_Reinhard2004",
    "tonemapping_operator_filmic",
]



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour_hdri.recovery.highlights` module."""

from __future__ import annotations

import numpy as np
import os
import platform
import re
import shlex
import shutil
import subprocess  # nosec
import tempfile
import unittest

from colour import read_image
from colour.hints import Boolean, List, NDArray

from colour_hdri import TESTS_RESOURCES_DIRECTORY
from colour_hdri.process import RAW_CONVERTER, RAW_D_CONVERSION_ARGUMENTS
from colour_hdri.recovery import (
    highlights_recovery_blend,
    highlights_recovery_LCHab,
)
from colour_hdri.models import camera_space_to_sRGB
from colour_hdri.utilities import filter_files

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "FROBISHER_001_DIRECTORY",
    "RECOVERY_DIRECTORY",
    "RAW_IMAGES",
    "XYZ_TO_CAMERA_SPACE_MATRIX",
    "TestHighlightsRecoveryBlend",
    "TestHighlightsRecoveryLCHab",
]

_IS_WINDOWS_PLATFORM: Boolean = platform.system() in ("Windows", "Microsoft")
"""Whether the current platform is *Windows*."""

FROBISHER_001_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "frobisher_001"
)

RECOVERY_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "colour_hdri", "recovery"
)

RAW_IMAGES: List[str] = filter_files(FROBISHER_001_DIRECTORY, ("CR2",))

XYZ_TO_CAMERA_SPACE_MATRIX: NDArray = np.array(
    [
        [0.47160000, 0.06030000, -0.08300000],
        [-0.77980000, 1.54740000, 0.24800000],
        [-0.14960000, 0.19370000, 0.66510000],
    ]
)


class TestHighlightsRecoveryBlend(unittest.TestCase):
    """
    Define :func:`colour_hdri.recovery.highlights.\
highlights_recovery_blend` definition unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._temporary_directory = tempfile.mkdtemp()

    def tearDown(self):
        """After tests actions."""

        shutil.rmtree(self._temporary_directory)

    def test_highlights_recovery_blend(self):
        """
        Test :func:`colour_hdri.recovery.highlights.highlights_recovery_blend`
        definition.
        """

        multipliers = np.array([2.42089718, 1.00000000, 1.54687415])
        multipliers /= np.max(multipliers)

        reference_raw_file = RAW_IMAGES[1]
        test_raw_file = os.path.join(
            self._temporary_directory, os.path.basename(reference_raw_file)
        )
        shutil.copyfile(reference_raw_file, test_raw_file)
        command = [RAW_CONVERTER] + shlex.split(
            RAW_D_CONVERSION_ARGUMENTS.format(test_raw_file),
            posix=not _IS_WINDOWS_PLATFORM,
        )

        subprocess.call(command)  # nosec

        test_tiff_file = read_image(
            str(re.sub("\\.CR2$", ".tiff", test_raw_file))
        )[::10, ::10, :]

        test_tiff_file *= multipliers
        test_tiff_file = highlights_recovery_blend(test_tiff_file, multipliers)
        test_tiff_file = camera_space_to_sRGB(
            test_tiff_file, XYZ_TO_CAMERA_SPACE_MATRIX
        )
        reference_exr_path = os.path.join(
            RECOVERY_DIRECTORY,
            os.path.basename(re.sub("\\.CR2$", "_Blend.exr", test_raw_file)),
        )
        reference_exr_file = read_image(str(reference_exr_path))

        np.testing.assert_allclose(
            test_tiff_file, reference_exr_file, rtol=0.0001, atol=0.0001
        )


class TestHighlightsRecoveryLCHab(unittest.TestCase):
    """
    Define :func:`colour_hdri.recovery.highlights.\
highlights_recovery_LCHab` definition unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._temporary_directory = tempfile.mkdtemp()

    def tearDown(self):
        """After tests actions."""

        shutil.rmtree(self._temporary_directory)

    def test_highlights_recovery_LCHab(self):
        """
        Test :func:`colour_hdri.recovery.highlights.highlights_recovery_LCHab`
        definition.
        """

        multipliers = np.array([2.42089718, 1.00000000, 1.54687415])
        multipliers /= np.max(multipliers)

        reference_raw_file = RAW_IMAGES[1]
        test_raw_file = os.path.join(
            self._temporary_directory, os.path.basename(reference_raw_file)
        )
        shutil.copyfile(reference_raw_file, test_raw_file)
        command = [RAW_CONVERTER] + shlex.split(
            RAW_D_CONVERSION_ARGUMENTS.format(test_raw_file),
            posix=not _IS_WINDOWS_PLATFORM,
        )

        subprocess.call(command)  # nosec

        test_tiff_file = read_image(
            str(re.sub("\\.CR2$", ".tiff", test_raw_file))
        )[::10, ::10, :]

        test_tiff_file *= multipliers
        test_tiff_file = highlights_recovery_LCHab(
            test_tiff_file, min(multipliers)
        )
        test_tiff_file = camera_space_to_sRGB(
            test_tiff_file, XYZ_TO_CAMERA_SPACE_MATRIX
        )

        reference_exr_path = os.path.join(
            RECOVERY_DIRECTORY,
            os.path.basename(re.sub("\\.CR2$", "_LCHab.exr", test_raw_file)),
        )
        reference_exr_file = read_image(str(reference_exr_path))

        np.testing.assert_allclose(
            test_tiff_file, reference_exr_file, rtol=0.0001, atol=0.0001
        )


if __name__ == "__main__":
    unittest.main()



from .common import (
    average_luminance,
    average_illuminance,
    luminance_to_exposure_value,
    illuminance_to_exposure_value,
    adjust_exposure,
)
from .dsc import (
    focal_plane_exposure,
    arithmetic_mean_focal_plane_exposure,
    saturation_based_speed_focal_plane_exposure,
    exposure_index_values,
    exposure_value_100,
    photometric_exposure_scale_factor_Lagarde2014,
)

__all__ = [
    "average_luminance",
    "average_illuminance",
    "luminance_to_exposure_value",
    "illuminance_to_exposure_value",
    "adjust_exposure",
]
__all__ += [
    "focal_plane_exposure",
    "arithmetic_mean_focal_plane_exposure",
    "saturation_based_speed_focal_plane_exposure",
    "exposure_index_values",
    "exposure_value_100",
    "photometric_exposure_scale_factor_Lagarde2014",
]



"""
Global Tonemapping Operators
============================

Defines the global tonemapping operators objects:

-   :func:`colour_hdri.tonemapping_operator_simple`
-   :func:`colour_hdri.tonemapping_operator_normalisation`
-   :func:`colour_hdri.tonemapping_operator_gamma`
-   :func:`colour_hdri.tonemapping_operator_logarithmic`
-   :func:`colour_hdri.tonemapping_operator_exponential`
-   :func:`colour_hdri.tonemapping_operator_logarithmic_mapping`
-   :func:`colour_hdri.tonemapping_operator_exponentiation_mapping`
-   :func:`colour_hdri.tonemapping_operator_Schlick1994`
-   :func:`colour_hdri.tonemapping_operator_Tumblin1999`
-   :func:`colour_hdri.tonemapping_operator_Reinhard2004`
-   :func:`colour_hdri.tonemapping_operator_filmic`

See Also
--------
`Colour - HDRI - Examples: Global Tonemapping Operators Jupyter Notebook
<https://github.com/colour-science/colour-hdri/blob/master/colour_hdri/\

examples/examples_global_tonemapping_operators.ipynb>`__

References
----------
-   :cite:`Banterle2011k` : Banterle, F., Artusi, A., Debattista, K., &
    Chalmers, A. (2011). 3.2.1 Simple Mapping Methods. In Advanced High
    Dynamic Range Imaging (pp. 38-41). A K Peters/CRC Press.
    ISBN:978-1-56881-719-4
-   :cite:`Habble2010d` : Habble, J. (2010). Filmic Tonemapping Operators.
    Retrieved March 15, 2015, from http://filmicgames.com/archives/75
-   :cite:`Habble2010e` : Habble, J. (2010). Uncharted 2: HDR Lighting.
    Retrieved March 15, 2015, from
    http://www.slideshare.net/ozlael/hable-john-uncharted2-hdr-lighting
-   :cite:`Reinhard2005c` : Reinhard, E., & Devlin, K. (2005). Dynamic Range
    Reduction Inspired by Photoreceptor Physiology. IEEE Transactions on
    Visualization and Computer Graphics, 11(01), 13-24. doi:10.1109/TVCG.2005.9
-   :cite:`Schlick1994` : Schlick, C. (1994). Quantization Techniques for
    Visualization of High Dynamic Range Pictures. Proceedings of the Fifth
    Eurographics Workshop on Rendering, Section 5, 7-18. ISSN:0920-5691
-   :cite:`Tumblin1999c` : Tumblin, J., Hodgins, J. K., & Guenter, B. K.
    (1999). Two methods for display of high contrast images. ACM Transactions
    on Graphics, 18(1), 56-94. doi:10.1145/300776.300783
-   :cite:`Wikipediabn` : Wikipedia. (n.d.). Tonemapping - Purpose and
    methods. Retrieved March 15, 2015, from
    http://en.wikipedia.org/wiki/Tone_mapping#Purpose_and_methods
"""

from __future__ import annotations

import numpy as np

from colour.constants import EPSILON
from colour.hints import (
    ArrayLike,
    Floating,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    NDArray,
)
from colour.models import RGB_COLOURSPACES, RGB_Colourspace, RGB_luminance
from colour.utilities import as_float_array

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "log_average",
    "tonemapping_operator_simple",
    "tonemapping_operator_normalisation",
    "tonemapping_operator_gamma",
    "tonemapping_operator_logarithmic",
    "tonemapping_operator_exponential",
    "tonemapping_operator_logarithmic_mapping",
    "tonemapping_operator_exponentiation_mapping",
    "tonemapping_operator_Schlick1994",
    "tonemapping_operator_Tumblin1999",
    "tonemapping_operator_Reinhard2004",
    "tonemapping_operator_filmic",
]


def log_average(
    a: FloatingOrArrayLike, epsilon: Floating = EPSILON
) -> FloatingOrNDArray:
    """
    Compute the log average of given array.

    Parameters
    ----------
    a
        Array to compute the log average.
    epsilon
        Constant to avoid singularities in computations.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Array log average.

    Examples
    --------
    >>> log_average(np.linspace(0, 10, 10))  # doctest: +ELLIPSIS
    0.1...
    """

    a = as_float_array(a)

    average = np.exp(np.average(np.log(a + epsilon)))

    return average


def tonemapping_operator_simple(RGB: ArrayLike) -> NDArray:
    """
    Perform given *RGB* array tonemapping using the simple method:
    :math:`\\cfrac{RGB}{RGB + 1}`.

    Parameters
    ----------
    RGB
        *RGB* array to perform tonemapping onto.

    Returns
    -------
    :class:`numpy.ndarray`
        Tonemapped *RGB* array.

    References
    ----------
    :cite:`Wikipediabn`

    Examples
    --------
    >>> tonemapping_operator_simple(np.array(
    ...     [[[0.48046875, 0.35156256, 0.23632812],
    ...       [1.39843753, 0.55468757, 0.39062594]],
    ...      [[4.40625388, 2.15625895, 1.34375372],
    ...       [6.59375023, 3.43751395, 2.21875829]]]))  # doctest: +ELLIPSIS
    array([[[ 0.3245382...,  0.2601156...,  0.1911532...],
            [ 0.5830618...,  0.3567839...,  0.2808993...]],
    <BLANKLINE>
           [[ 0.8150290...,  0.6831692...,  0.5733340...],
            [ 0.8683127...,  0.7746486...,  0.6893211...]]])
    """

    RGB = as_float_array(RGB)

    return RGB / (RGB + 1)


def tonemapping_operator_normalisation(
    RGB: ArrayLike, colourspace: RGB_Colourspace = RGB_COLOURSPACES["sRGB"]
) -> NDArray:
    """
    Perform given *RGB* array tonemapping using the normalisation method.

    Parameters
    ----------
    RGB
        *RGB* array to perform tonemapping onto.
    colourspace
        *RGB* colourspace used for internal *Luminance* computation.

    Returns
    -------
    :class:`numpy.ndarray`
        Tonemapped *RGB* array.

    References
    ----------
    :cite:`Banterle2011k`

    Examples
    --------
    >>> tonemapping_operator_normalisation(np.array(
    ...     [[[0.48046875, 0.35156256, 0.23632812],
    ...       [1.39843753, 0.55468757, 0.39062594]],
    ...      [[4.40625388, 2.15625895, 1.34375372],
    ...       [6.59375023, 3.43751395, 2.21875829]]]))  # doctest: +ELLIPSIS
    array([[[ 0.1194997...,  0.0874388...,  0.0587783...],
            [ 0.3478122...,  0.1379590...,  0.0971544...]],
    <BLANKLINE>
           [[ 1.0959009...,  0.5362936...,  0.3342115...],
            [ 1.6399638...,  0.8549608...,  0.5518382...]]])
    """

    RGB = as_float_array(RGB)

    L = RGB_luminance(RGB, colourspace.primaries, colourspace.whitepoint)
    L_max = as_float_array(np.max(L))

    return RGB / L_max


def tonemapping_operator_gamma(
    RGB: ArrayLike, gamma: Floating = 1, EV: Floating = 0
) -> NDArray:
    """
    Perform given *RGB* array tonemapping using the gamma and exposure
    correction method.

    Parameters
    ----------
    RGB
        *RGB* array to perform tonemapping onto.
    gamma
        :math:`\\gamma` correction value.
    EV
        Exposure adjustment value.

    Returns
    -------
    :class:`numpy.ndarray`
        Tonemapped *RGB* array.

    References
    ----------
    :cite:`Banterle2011k`

    Examples
    --------
    >>> tonemapping_operator_gamma(np.array(
    ...     [[[0.48046875, 0.35156256, 0.23632812],
    ...       [1.39843753, 0.55468757, 0.39062594]],
    ...      [[4.40625388, 2.15625895, 1.34375372],
    ...       [6.59375023, 3.43751395, 2.21875829]]]),
    ...      1.0, -3.0)  # doctest: +ELLIPSIS
    array([[[ 0.0600585...,  0.0439453...,  0.0295410...],
            [ 0.1748046...,  0.0693359...,  0.0488282...]],
    <BLANKLINE>
           [[ 0.5507817...,  0.2695323...,  0.1679692...],
            [ 0.8242187...,  0.4296892...,  0.2773447...]]])
    """

    RGB = as_float_array(RGB)

    exposure = 2**EV
    RGB = (exposure * RGB) ** (1 / gamma)

    return RGB


def tonemapping_operator_logarithmic(
    RGB: ArrayLike,
    q: Floating = 1,
    k: Floating = 1,
    colourspace: RGB_Colourspace = RGB_COLOURSPACES["sRGB"],
) -> NDArray:
    """
    Perform given *RGB* array tonemapping using the logarithmic method.

    Parameters
    ----------
    RGB
        *RGB* array to perform tonemapping onto.
    q
        :math:`q`.
    k
        :math:`k`.
    colourspace
        *RGB* colourspace used for internal *Luminance* computation.

    Returns
    -------
    :class:`numpy.ndarray`
        Tonemapped *RGB* array.

    References
    ----------
    :cite:`Banterle2011k`

    Examples
    --------
    >>> tonemapping_operator_logarithmic(np.array(
    ...     [[[0.48046875, 0.35156256, 0.23632812],
    ...       [1.39843753, 0.55468757, 0.39062594]],
    ...      [[4.40625388, 2.15625895, 1.34375372],
    ...       [6.59375023, 3.43751395, 2.21875829]]]),
    ...       1.0, 25)  # doctest: +ELLIPSIS
    array([[[ 0.0884587...,  0.0647259...,  0.0435102...],
            [ 0.2278222...,  0.0903652...,  0.0636376...]],
    <BLANKLINE>
           [[ 0.4717487...,  0.2308565...,  0.1438669...],
            [ 0.5727396...,  0.2985858...,  0.1927235...]]])
    """

    RGB = as_float_array(RGB)

    q = 1 if q < 1 else q
    k = 1 if k < 1 else k

    L = as_float_array(
        RGB_luminance(RGB, colourspace.primaries, colourspace.whitepoint)
    )
    L_max = np.max(L)
    L_d = as_float_array(np.log10(1 + L * q) / np.log10(1 + L_max * k))

    return RGB * L_d[..., np.newaxis] / L[..., np.newaxis]


def tonemapping_operator_exponential(
    RGB: ArrayLike,
    q: Floating = 1,
    k: Floating = 1,
    colourspace: RGB_Colourspace = RGB_COLOURSPACES["sRGB"],
) -> NDArray:
    """
    Perform given *RGB* array tonemapping using the exponential method.

    Parameters
    ----------
    RGB
        *RGB* array to perform tonemapping onto.
    q
        :math:`q`.
    k
        :math:`k`.
    colourspace
        *RGB* colourspace used for internal *Luminance* computation.

    Returns
    -------
    :class:`numpy.ndarray`
        Tonemapped *RGB* array.

    References
    ----------
    :cite:`Banterle2011k`

    Examples
    --------
    >>> tonemapping_operator_exponential(np.array(
    ...     [[[0.48046875, 0.35156256, 0.23632812],
    ...       [1.39843753, 0.55468757, 0.39062594]],
    ...      [[4.40625388, 2.15625895, 1.34375372],
    ...       [6.59375023, 3.43751395, 2.21875829]]]),
    ...       1.0, 25)  # doctest: +ELLIPSIS
    array([[[ 0.0148082...,  0.0108353...,  0.0072837...],
            [ 0.0428669...,  0.0170031...,  0.0119740...]],
    <BLANKLINE>
           [[ 0.1312736...,  0.0642404...,  0.0400338...],
            [ 0.1921684...,  0.1001830...,  0.0646635...]]])
    """

    RGB = as_float_array(RGB)

    q = 1 if q < 1 else q
    k = 1 if k < 1 else k

    L = as_float_array(
        RGB_luminance(RGB, colourspace.primaries, colourspace.whitepoint)
    )
    L_a = log_average(L)
    L_d = as_float_array(1 - np.exp(-(L * q) / (L_a * k)))

    return RGB * L_d[..., np.newaxis] / L[..., np.newaxis]


def tonemapping_operator_logarithmic_mapping(
    RGB: ArrayLike,
    p: Floating = 1,
    q: Floating = 1,
    colourspace: RGB_Colourspace = RGB_COLOURSPACES["sRGB"],
) -> NDArray:
    """
    Perform given *RGB* array tonemapping using the logarithmic mapping
    method.

    Parameters
    ----------
    RGB
        *RGB* array to perform tonemapping onto.
    p
        :math:`p`.
    q
        :math:`q`.
    colourspace
        *RGB* colourspace used for internal *Luminance* computation.

    Returns
    -------
    :class:`numpy.ndarray`
        Tonemapped *RGB* array.

    References
    ----------
    :cite:`Schlick1994`

    Examples
    --------
    >>> tonemapping_operator_logarithmic_mapping(np.array(
    ...     [[[0.48046875, 0.35156256, 0.23632812],
    ...       [1.39843753, 0.55468757, 0.39062594]],
    ...      [[4.40625388, 2.15625895, 1.34375372],
    ...       [6.59375023, 3.43751395, 2.21875829]]]))  # doctest: +ELLIPSIS
    array([[[ 0.2532899...,  0.1853341...,  0.1245857...],
            [ 0.6523387...,  0.2587489...,  0.1822179...]],
    <BLANKLINE>
           [[ 1.3507897...,  0.6610269...,  0.4119437...],
            [ 1.6399638...,  0.8549608...,  0.5518382...]]])
    """

    RGB = as_float_array(RGB)

    L = as_float_array(
        RGB_luminance(RGB, colourspace.primaries, colourspace.whitepoint)
    )
    L_max = np.max(L)
    L_d = as_float_array(
        (np.log(1 + p * L) / np.log(1 + p * L_max)) ** (1 / q)
    )

    return RGB * L_d[..., np.newaxis] / L[..., np.newaxis]


def tonemapping_operator_exponentiation_mapping(
    RGB: ArrayLike,
    p: Floating = 1,
    q: Floating = 1,
    colourspace: RGB_Colourspace = RGB_COLOURSPACES["sRGB"],
) -> NDArray:
    """
    Perform given *RGB* array tonemapping using the exponentiation mapping
    method.

    Parameters
    ----------
    RGB
        *RGB* array to perform tonemapping onto.
    p
        :math:`p`.
    q
        :math:`q`.
    colourspace
        *RGB* colourspace used for internal *Luminance* computation.

    Returns
    -------
    :class:`numpy.ndarray`
        Tonemapped *RGB* array.

    References
    ----------
    :cite:`Schlick1994`

    Examples
    --------
    >>> tonemapping_operator_exponentiation_mapping(np.array(
    ...     [[[0.48046875, 0.35156256, 0.23632812],
    ...       [1.39843753, 0.55468757, 0.39062594]],
    ...      [[4.40625388, 2.15625895, 1.34375372],
    ...       [6.59375023, 3.43751395, 2.21875829]]]))  # doctest: +ELLIPSIS
    array([[[ 0.1194997...,  0.0874388...,  0.0587783...],
            [ 0.3478122...,  0.1379590...,  0.0971544...]],
    <BLANKLINE>
           [[ 1.0959009...,  0.5362936...,  0.3342115...],
            [ 1.6399638...,  0.8549608...,  0.5518382...]]])
    """

    RGB = as_float_array(RGB)

    L = as_float_array(
        RGB_luminance(RGB, colourspace.primaries, colourspace.whitepoint)
    )
    L_max = np.max(L)
    L_d = as_float_array((L / L_max) ** (p / q))

    return RGB * L_d[..., np.newaxis] / L[..., np.newaxis]


def tonemapping_operator_Schlick1994(
    RGB: ArrayLike,
    p: Floating = 1,
    colourspace: RGB_Colourspace = RGB_COLOURSPACES["sRGB"],
) -> NDArray:
    """
    Perform given *RGB* array tonemapping using *Schlick (1994)* method.

    Parameters
    ----------
    RGB
        *RGB* array to perform tonemapping onto.
    p
        :math:`p`.
    colourspace
        *RGB* colourspace used for internal *Luminance* computation.

    Returns
    -------
    :class:`numpy.ndarray`
        Tonemapped *RGB* array.

    References
    ----------
    :cite:`Banterle2011k`, :cite:`Schlick1994`

    Examples
    --------
    >>> tonemapping_operator_Schlick1994(np.array(
    ...     [[[0.48046875, 0.35156256, 0.23632812],
    ...       [1.39843753, 0.55468757, 0.39062594]],
    ...      [[4.40625388, 2.15625895, 1.34375372],
    ...       [6.59375023, 3.43751395, 2.21875829]]]))  # doctest: +ELLIPSIS
    array([[[ 0.1194997...,  0.0874388...,  0.0587783...],
            [ 0.3478122...,  0.1379590...,  0.0971544...]],
    <BLANKLINE>
           [[ 1.0959009...,  0.5362936...,  0.3342115...],
            [ 1.6399638...,  0.8549608...,  0.5518382...]]])
    """

    # TODO: Implement automatic *p* and *non-uniform* computations support.

    RGB = as_float_array(RGB)

    L = as_float_array(
        RGB_luminance(RGB, colourspace.primaries, colourspace.whitepoint)
    )
    L_max = np.max(L)
    L_d = as_float_array((p * L) / (p * L - L + L_max))

    return RGB * L_d[..., np.newaxis] / L[..., np.newaxis]


def tonemapping_operator_Tumblin1999(
    RGB: ArrayLike,
    L_da: Floating = 20,
    C_max: Floating = 100,
    L_max: Floating = 100,
    colourspace: RGB_Colourspace = RGB_COLOURSPACES["sRGB"],
) -> NDArray:
    """
    Perform given *RGB* array tonemapping using
    *Tumblin, Hodgins and Guenter (1999)* method.

    Parameters
    ----------
    RGB
        *RGB* array to perform tonemapping onto.
    L_da
        :math:`L_{da}` display adaptation luminance, a mid-range display value.
    C_max
        :math:`C_{max}` maximum contrast available from the display.
    L_max
        :math:`L_{max}` maximum display luminance.
    colourspace
        *RGB* colourspace used for internal *Luminance* computation.

    Returns
    -------
    :class:`numpy.ndarray`
        Tonemapped *RGB* array.

    References
    ----------
    :cite:`Tumblin1999c`

    Examples
    --------
    >>> tonemapping_operator_Tumblin1999(np.array(
    ...     [[[0.48046875, 0.35156256, 0.23632812],
    ...       [1.39843753, 0.55468757, 0.39062594]],
    ...      [[4.40625388, 2.15625895, 1.34375372],
    ...       [6.59375023, 3.43751395, 2.21875829]]]))  # doctest: +ELLIPSIS
    array([[[ 0.0400492...,  0.0293043...,  0.0196990...],
            [ 0.1019768...,  0.0404489...,  0.0284852...]],
    <BLANKLINE>
           [[ 0.2490212...,  0.1218618...,  0.0759427...],
            [ 0.3408366...,  0.1776880...,  0.1146895...]]])
    """

    RGB = as_float_array(RGB)

    L_w = as_float_array(
        RGB_luminance(RGB, colourspace.primaries, colourspace.whitepoint)
    )

    def f(x: FloatingOrNDArray) -> FloatingOrNDArray:
        return np.where(
            x > 100, 2.655, 1.855 + 0.4 * np.log10(x + 2.3 * 10**-5)
        )

    L_wa = np.exp(np.mean(np.log(L_w + 2.3 * 10**-5)))
    g_d = f(L_da)
    g_w = f(L_wa)
    g_wd = g_w / (1.855 + 0.4 * np.log(L_da))

    mL_wa = np.sqrt(C_max) ** (g_wd - 1)

    L_d = mL_wa * L_da * (L_w / L_wa) ** (g_w / g_d)

    return (RGB * L_d[..., np.newaxis] / L_w[..., np.newaxis]) / L_max


def tonemapping_operator_Reinhard2004(
    RGB: ArrayLike,
    f: Floating = 0,
    m: Floating = 0.3,
    a: Floating = 0,
    c: Floating = 0,
    colourspace: RGB_Colourspace = RGB_COLOURSPACES["sRGB"],
) -> NDArray:
    """
    Perform given *RGB* array tonemapping using *Reinhard and Devlin (2004)*
    method.

    Parameters
    ----------
    RGB
        *RGB* array to perform tonemapping onto.
    f
        :math:`f`.
    m
        :math:`m`.
    a
        :math:`a`.
    c
        :math:`c`.
    colourspace
        *RGB* colourspace used for internal *Luminance* computation.

    Returns
    -------
    :class:`numpy.ndarray`
        Tonemapped *RGB* array.

    References
    ----------
    :cite:`Reinhard2005c`

    Examples
    --------
    >>> tonemapping_operator_Reinhard2004(np.array(
    ...     [[[0.48046875, 0.35156256, 0.23632812],
    ...       [1.39843753, 0.55468757, 0.39062594]],
    ...      [[4.40625388, 2.15625895, 1.34375372],
    ...       [6.59375023, 3.43751395, 2.21875829]]]),
    ...     -10)  # doctest: +ELLIPSIS
    array([[[ 0.0216792...,  0.0159556...,  0.0107821...],
            [ 0.0605894...,  0.0249445...,  0.0176972...]],
    <BLANKLINE>
           [[ 0.1688972...,  0.0904532...,  0.0583584...],
            [ 0.2331935...,  0.1368456...,  0.0928316...]]])
    """

    RGB = as_float_array(RGB)

    C_av = np.array(
        (
            np.average(RGB[..., 0]),
            np.average(RGB[..., 1]),
            np.average(RGB[..., 2]),
        )
    )

    L = as_float_array(
        RGB_luminance(RGB, colourspace.primaries, colourspace.whitepoint)
    )

    L_lav = log_average(L)
    L_min, L_max = np.min(L), np.max(L)

    f = np.exp(-f)

    m = (
        m
        if m > 0
        else (
            0.3
            + 0.7
            * (
                (np.log(L_max) - L_lav)
                / (np.log(L_max) - np.log(L_min)) ** 1.4
            )
        )
    )

    I_l = (c * RGB + (1 - c)) * L[..., np.newaxis]
    I_g = c * C_av + (1 - c) * L_lav
    I_a = a * I_l + (1 - a) * I_g

    return RGB / (RGB + (f * I_a) ** m)


def tonemapping_operator_filmic(
    RGB: ArrayLike,
    shoulder_strength: Floating = 0.22,
    linear_strength: Floating = 0.3,
    linear_angle: Floating = 0.1,
    toe_strength: Floating = 0.2,
    toe_numerator: Floating = 0.01,
    toe_denominator: Floating = 0.3,
    exposure_bias: Floating = 2,
    linear_whitepoint: Floating = 11.2,
) -> NDArray:
    """
    Perform given *RGB* array tonemapping using *Habble (2010)* method.

    Parameters
    ----------
    RGB
        *RGB* array to perform tonemapping onto.
    shoulder_strength
        Shoulder strength.
    linear_strength
        Linear strength.
    linear_angle
        Linear angle.
    toe_strength
        Toe strength.
    toe_numerator
        Toe numerator.
    toe_denominator
        Toe denominator.
    exposure_bias
        Exposure bias.
    linear_whitepoint
        Linear whitepoint.

    Returns
    -------
    :class:`numpy.ndarray`
        Tonemapped *RGB* array.

    References
    ----------
    :cite:`Habble2010d`, :cite:`Habble2010e`

    Examples
    --------
    >>> tonemapping_operator_filmic(np.array(
    ...     [[[0.48046875, 0.35156256, 0.23632812],
    ...       [1.39843753, 0.55468757, 0.39062594]],
    ...      [[4.40625388, 2.15625895, 1.34375372],
    ...       [6.59375023, 3.43751395, 2.21875829]]]))  # doctest: +ELLIPSIS
    array([[[ 0.4507954...,  0.3619673...,  0.2617269...],
            [ 0.7567191...,  0.4933310...,  0.3911730...]],
    <BLANKLINE>
           [[ 0.9725554...,  0.8557374...,  0.7465713...],
            [ 1.0158782...,  0.9382937...,  0.8615161...]]])
    """

    RGB = as_float_array(RGB)

    A = shoulder_strength
    B = linear_strength
    C = linear_angle
    D = toe_strength
    E = toe_numerator
    F = toe_denominator

    def f(x: FloatingOrNDArray, A, B, C, D, E, F) -> FloatingOrNDArray:
        return (
            (x * (A * x + C * B) + D * E) / (x * (A * x + B) + D * F)
        ) - E / F

    RGB = f(RGB * exposure_bias, A, B, C, D, E, F)

    return RGB * (1 / as_float_array(f(linear_whitepoint, A, B, C, D, E, F)))



"""
Common Utilities
================

Defines the common utilities objects that don't fall in any specific category.
"""

from __future__ import annotations

import os
import re
from collections import defaultdict

from colour.hints import Boolean, Dict, List, Optional, Sequence, Union

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "vivification",
    "vivified_to_dict",
    "path_exists",
    "filter_files",
]


def vivification() -> defaultdict:
    """
    Implement supports for vivification of the underlying dict like
    data-structure, magical!

    Returns
    -------
    :class:`defaultdict`

    Examples
    --------
    >>> vivified = vivification()
    >>> vivified['my']['attribute'] = 1
    >>> vivified['my']  # doctest: +SKIP
    defaultdict(<function vivification at 0x...>, {u'attribute': 1})
    >>> vivified['my']['attribute']
    1
    """

    return defaultdict(vivification)


def vivified_to_dict(vivified: Union[Dict, defaultdict]) -> Dict:
    """
    Convert given vivified data-structure to dictionary.

    Parameters
    ----------
    vivified
        Vivified data-structure.

    Returns
    -------
    :class:`dict`

    Examples
    --------
    >>> vivified = vivification()
    >>> vivified['my']['attribute'] = 1
    >>> vivified_to_dict(vivified)  # doctest: +SKIP
    {u'my': {u'attribute': 1}}
    """

    if isinstance(vivified, defaultdict):
        vivified = {
            key: vivified_to_dict(value) for key, value in vivified.items()
        }
    return vivified


def path_exists(path: Optional[str]) -> Boolean:
    """
    Return whether given path exists.

    Parameters
    ----------
    path
        Path to check the existence.

    Returns
    -------
    :class:`bool`
        Whether given path exists.

    Examples
    --------
    >>> path_exists(__file__)
    True
    >>> path_exists('')
    False
    """

    if path is None:
        return False
    else:
        return os.path.exists(path)


def filter_files(directory: str, extensions: Sequence[str]) -> List[str]:
    """
    Filter given directory for files matching given extensions.

    Parameters
    ----------
    directory
        Directory to filter.
    extensions
        Extensions to filter on.

    Returns
    -------
    :class:`list`
        Filtered files.
    """

    return [
        os.path.join(directory, path)
        for path in filter(
            lambda x: re.search(f"{'|'.join(extensions)}$", x),
            sorted(os.listdir(directory)),
        )
    ]



from .datasets import *  # noqa
from . import datasets
from .dng import (
    xy_to_camera_neutral,
    camera_neutral_to_xy,
    XYZ_to_camera_space_matrix,
    camera_space_to_XYZ_matrix,
)
from .rgb import camera_space_to_RGB, camera_space_to_sRGB

__all__ = []
__all__ += datasets.__all__
__all__ += [
    "xy_to_camera_neutral",
    "camera_neutral_to_xy",
    "XYZ_to_camera_space_matrix",
    "camera_space_to_XYZ_matrix",
]
__all__ += [
    "camera_space_to_RGB",
    "camera_space_to_sRGB",
]



# !/usr/bin/env python
"""
Defines the unit tests for the :mod:`colour_hdri.models.dng` module.

Notes
-----
-   The various matrices used in this module are extracted from a
    *Canon EOS 5D Mark II* camera.
"""

from __future__ import annotations

import numpy as np
import unittest

from colour.hints import Floating, NDArray

from colour_hdri.models.dng import interpolated_matrix
from colour_hdri.models import (
    xy_to_camera_neutral,
    camera_neutral_to_xy,
    XYZ_to_camera_space_matrix,
    camera_space_to_XYZ_matrix,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CCT_CALIBRATION_ILLUMINANT_1",
    "CCT_CALIBRATION_ILLUMINANT_2",
    "M_COLOR_MATRIX_1",
    "M_COLOR_MATRIX_2",
    "M_CAMERA_CALIBRATION_1",
    "M_CAMERA_CALIBRATION_2",
    "ANALOG_BALANCE",
    "M_FORWARD_MATRIX_1",
    "M_FORWARD_MATRIX_2",
    "TestInterpolatedMatrix",
    "TestXy_to_camera_neutral",
    "TestCamera_neutral_to_xy",
    "TestXYZ_to_camera_space_matrix",
    "TestCamera_space_to_XYZ_matrix",
]

CCT_CALIBRATION_ILLUMINANT_1: Floating = 2850
CCT_CALIBRATION_ILLUMINANT_2: Floating = 6500

M_COLOR_MATRIX_1: NDArray = np.array(
    [
        [0.5309, -0.0229, -0.0336],
        [-0.6241, 1.3265, 0.3337],
        [-0.0817, 0.1215, 0.6664],
    ]
)
M_COLOR_MATRIX_2: NDArray = np.array(
    [
        [0.4716, 0.0603, -0.0830],
        [-0.7798, 1.5474, 0.2480],
        [-0.1496, 0.1937, 0.6651],
    ]
)

M_CAMERA_CALIBRATION_1: NDArray = np.array(
    [
        [0.9603, 0.0000, 0.0000],
        [0.0000, 1.0000, 0.0000],
        [0.0000, 0.0000, 0.9664],
    ]
)
M_CAMERA_CALIBRATION_2: NDArray = np.array(
    [
        [0.9603, 0.0000, 0.0000],
        [0.0000, 1.0000, 0.0000],
        [0.0000, 0.0000, 0.9664],
    ]
)

ANALOG_BALANCE: NDArray = np.ones(3)

M_FORWARD_MATRIX_1: NDArray = np.array(
    [
        [0.8924, -0.1041, 0.1760],
        [0.4351, 0.6621, -0.0972],
        [0.0505, -0.1562, 0.9308],
    ]
)
M_FORWARD_MATRIX_2: NDArray = np.array(
    [
        [0.8924, -0.1041, 0.1760],
        [0.4351, 0.6621, -0.0972],
        [0.0505, -0.1562, 0.9308],
    ]
)


class TestInterpolatedMatrix(unittest.TestCase):
    """
    Define :func:`colour_hdri.models.adobe_dng.\
interpolated_matrix` definition unit tests methods.
    """

    def test_interpolated_matrix(self):
        """
        Test :func:`colour_hdri.models.adobe_dng.\
interpolated_matrix` definition.
        """

        M_reference = np.array(
            [
                [0.48549082, 0.04081068, -0.07142822],
                [-0.74332781, 1.49565493, 0.26807493],
                [-0.13369466, 0.17678740, 0.66540452],
            ]
        )
        np.testing.assert_almost_equal(
            interpolated_matrix(
                5000, 2850, 6500, M_COLOR_MATRIX_1, M_COLOR_MATRIX_2
            ),
            M_reference,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            interpolated_matrix(
                2850, 2850, 6500, M_COLOR_MATRIX_1, M_COLOR_MATRIX_2
            ),
            M_COLOR_MATRIX_1,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            interpolated_matrix(
                1000, 2850, 6500, M_COLOR_MATRIX_1, M_COLOR_MATRIX_2
            ),
            M_COLOR_MATRIX_1,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            interpolated_matrix(
                6500, 2850, 6500, M_COLOR_MATRIX_1, M_COLOR_MATRIX_2
            ),
            M_COLOR_MATRIX_2,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            interpolated_matrix(
                10000, 2850, 6500, M_COLOR_MATRIX_1, M_COLOR_MATRIX_2
            ),
            M_COLOR_MATRIX_2,
            decimal=7,
        )


class TestXy_to_camera_neutral(unittest.TestCase):
    """
    Define :func:`colour_hdri.models.adobe_dng.\
xy_to_camera_neutral` definition unit tests methods.
    """

    def test_xy_to_camera_neutral(self):
        """
        Test :func:`colour_hdri.models.adobe_dng.\
xy_to_camera_neutral` definition.
        """

        np.testing.assert_almost_equal(
            xy_to_camera_neutral(
                np.array([0.32816244, 0.34698169]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                M_COLOR_MATRIX_2,
                np.identity(3),
                np.identity(3),
                ANALOG_BALANCE,
            ),
            np.array([0.41306999, 1.00000000, 0.64646500]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            xy_to_camera_neutral(
                np.array([0.32816244, 0.34698169]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                np.identity(3),
                np.identity(3),
                np.identity(3),
                ANALOG_BALANCE,
            ),
            np.array([0.42696744, 1.00000000, 0.63712786]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            xy_to_camera_neutral(
                np.array([0.32816244, 0.34698169]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                M_COLOR_MATRIX_2,
                M_CAMERA_CALIBRATION_1,
                M_CAMERA_CALIBRATION_2,
                ANALOG_BALANCE,
            ),
            np.array([0.39667111, 1.00000000, 0.62474378]),
            decimal=7,
        )


class TestCamera_neutral_to_xy(unittest.TestCase):
    """
    Define :func:`colour_hdri.models.adobe_dng.\
camera_neutral_to_xy` definition unit tests methods.
    """

    def test_camera_neutral_to_xy(self):
        """
        Test :func:`colour_hdri.models.adobe_dng.\
camera_neutral_to_xy` definition.
        """

        np.testing.assert_almost_equal(
            camera_neutral_to_xy(
                np.array([0.41306999, 1.00000000, 0.64646500]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                M_COLOR_MATRIX_2,
                np.identity(3),
                np.identity(3),
                ANALOG_BALANCE,
            ),
            np.array([0.32816244, 0.34698169]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            camera_neutral_to_xy(
                np.array([0.42696744, 1.00000000, 0.63712786]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                np.identity(3),
                np.identity(3),
                np.identity(3),
                ANALOG_BALANCE,
            ),
            np.array([0.32816244, 0.34698169]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            camera_neutral_to_xy(
                np.array([0.39667111, 1.00000000, 0.62474378]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                M_COLOR_MATRIX_2,
                M_CAMERA_CALIBRATION_1,
                M_CAMERA_CALIBRATION_2,
                ANALOG_BALANCE,
            ),
            np.array([0.32816244, 0.34698169]),
            decimal=7,
        )


class TestXYZ_to_camera_space_matrix(unittest.TestCase):
    """
    Define :func:`colour_hdri.models.adobe_dng.\
XYZ_to_camera_space_matrix` definition unit tests methods.
    """

    def test_XYZ_to_camera_space_matrix(self):
        """
        Test :func:`colour_hdri.models.adobe_dng.\
XYZ_to_camera_space_matrix` definition.
        """
        M_reference = np.array(
            [
                [0.47823951, 0.05098453, -0.07746894],
                [-0.76236708, 1.52266698, 0.25759538],
                [-0.14199759, 0.18561614, 0.66524555],
            ]
        )
        np.testing.assert_almost_equal(
            XYZ_to_camera_space_matrix(
                np.array([0.32816244, 0.34698169]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                M_COLOR_MATRIX_2,
                np.identity(3),
                np.identity(3),
                ANALOG_BALANCE,
            ),
            M_reference,
            decimal=7,
        )

        M_reference = np.array(
            [
                [0.47823951, 0.05098453, -0.07746894],
                [-0.76236708, 1.52266698, 0.25759538],
                [-0.14199759, 0.18561614, 0.66524555],
            ]
        )
        np.testing.assert_almost_equal(
            XYZ_to_camera_space_matrix(
                np.array([0.32816244, 0.34698169]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                M_COLOR_MATRIX_2,
                np.identity(3),
                np.identity(3),
                ANALOG_BALANCE,
            ),
            M_reference,
            decimal=7,
        )

        M_reference = np.array(
            [
                [0.47823951, 0.05098453, -0.07746894],
                [-0.76236708, 1.52266698, 0.25759538],
                [-0.14199759, 0.18561614, 0.66524555],
            ]
        )
        np.testing.assert_almost_equal(
            XYZ_to_camera_space_matrix(
                np.array([0.32816244, 0.34698169]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                M_COLOR_MATRIX_2,
                np.identity(3),
                np.identity(3),
                ANALOG_BALANCE,
            ),
            M_reference,
            decimal=7,
        )


class TestCamera_space_to_XYZ_matrix(unittest.TestCase):
    """
    Define :func:`colour_hdri.models.adobe_dng.\
camera_space_to_XYZ_matrix` definition unit tests methods.
    """

    def test_camera_space_to_XYZ_matrix(self):
        """
        Test :func:`colour_hdri.models.adobe_dng.\
camera_space_to_XYZ_matrix` definition.
        """

        M_reference = np.array(
            [
                [2.16040871, -0.10410000, 0.27224985],
                [1.05333240, 0.66210000, -0.15035617],
                [0.12225531, -0.15620000, 1.43983046],
            ]
        )
        np.testing.assert_almost_equal(
            camera_space_to_XYZ_matrix(
                np.array([0.32816244, 0.34698169]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                M_COLOR_MATRIX_2,
                np.identity(3),
                np.identity(3),
                ANALOG_BALANCE,
                M_FORWARD_MATRIX_1,
                M_FORWARD_MATRIX_2,
            ),
            M_reference,
            decimal=7,
        )

        M_reference = np.array(
            [
                [2.16040871, -0.10410000, 0.27224985],
                [1.05333240, 0.66210000, -0.15035617],
                [0.12225531, -0.15620000, 1.43983046],
            ]
        )
        np.testing.assert_almost_equal(
            camera_space_to_XYZ_matrix(
                np.array([0.32816244, 0.34698169]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                M_COLOR_MATRIX_2,
                np.identity(3),
                np.identity(3),
                ANALOG_BALANCE,
                M_FORWARD_MATRIX_1,
                M_FORWARD_MATRIX_2,
            ),
            M_reference,
            decimal=7,
        )

        M_reference = np.array(
            [
                [2.16040871, -0.10410000, 0.27224985],
                [1.05333240, 0.66210000, -0.15035617],
                [0.12225531, -0.15620000, 1.43983046],
            ]
        )
        np.testing.assert_almost_equal(
            camera_space_to_XYZ_matrix(
                np.array([0.32816244, 0.34698169]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                M_COLOR_MATRIX_2,
                np.identity(3),
                np.identity(3),
                ANALOG_BALANCE,
                M_FORWARD_MATRIX_1,
                M_FORWARD_MATRIX_2,
            ),
            M_reference,
            decimal=7,
        )

        M_reference = np.array(
            [
                [2.08092549, -0.08875351, 0.23804475],
                [1.00983630, 0.63301091, -0.14108066],
                [0.13472434, -0.17097714, 1.40232276],
            ]
        )
        np.testing.assert_almost_equal(
            camera_space_to_XYZ_matrix(
                np.array([0.32816244, 0.34698169]),
                2850,
                6500,
                M_COLOR_MATRIX_1,
                M_COLOR_MATRIX_2,
                np.identity(3),
                np.identity(3),
                ANALOG_BALANCE,
                np.identity(3),
                np.identity(3),
            ),
            M_reference,
            decimal=7,
        )


if __name__ == "__main__":
    unittest.main()



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour_hdri.exposure.dsc` module."""

import numpy as np
import unittest

from colour_hdri.exposure.dsc import q_factor
from colour_hdri.exposure import (
    focal_plane_exposure,
    arithmetic_mean_focal_plane_exposure,
    saturation_based_speed_focal_plane_exposure,
    exposure_index_values,
    exposure_value_100,
    photometric_exposure_scale_factor_Lagarde2014,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestQFactor",
    "TestQFactor",
    "TestFocalPlaneExposure",
    "TestArithmeticMeanFocalPlaneExposure",
    "TestSaturationBasedSpeedFocalPlaneExposure",
    "TestExposureValue100",
    "TestPhotometricExposureScaleFactorLagarde2014",
]


class TestQFactor(unittest.TestCase):
    """
    Define :func:`colour_hdri.exposure.dsc.q_factor` definition
    unit tests methods.
    """

    def test_q_factor(self):
        """Test :func:`colour_hdri.exposure.dsc.q_factor` definition."""

        np.testing.assert_almost_equal(
            q_factor(
                np.array([9 / 10, 8 / 10, 9 / 10]),
                np.array([98 / 100, 98 / 100, 94 / 100]),
                np.array([10, 10, 20]),
            ),
            np.array([0.65157483, 0.57917763, 0.51808797]),
            decimal=7,
        )


class TestFocalPlaneExposure(unittest.TestCase):
    """
    Define :func:`colour_hdri.exposure.dsc.focal_plane_exposure` definition
    unit tests methods.
    """

    def test_focal_plane_exposure(self):
        """Test :func:`colour_hdri.exposure.dsc.focal_plane_exposure` definition."""

        np.testing.assert_almost_equal(
            focal_plane_exposure(
                np.array([4000, 2000, 1000]),
                np.array([8, 5.6, 2.8]),
                np.array([1 / 250, 1 / 500, 1 / 1000]),
                np.array([50 / 1000, 100 / 1000, 500 / 1000]),
                np.array([50 / 1000, 100 / 1000, 500 / 1000]),
                np.array([0.0015, 0.0050, 0.0100]),
                9 / 10,
                98 / 100,
                10,
            ),
            np.array([0.16439371, 0.08810904, 0.09310904]),
            decimal=7,
        )


class TestArithmeticMeanFocalPlaneExposure(unittest.TestCase):
    """
    Define :func:`colour_hdri.exposure.dsc.\
arithmetic_mean_focal_plane_exposure` definition unit tests methods.
    """

    def test_arithmetic_mean_focal_plane_exposure(self):
        """
        Test :func:`colour_hdri.exposure.dsc.\
arithmetic_mean_focal_plane_exposure` definition.
        """

        np.testing.assert_almost_equal(
            arithmetic_mean_focal_plane_exposure(
                np.array([4000, 2000, 1000]),
                np.array([8, 5.6, 2.8]),
                np.array([1 / 250, 1 / 500, 1 / 1000]),
            ),
            np.array([0.16289371, 0.08310904, 0.08310904]),
            decimal=7,
        )


class TestSaturationBasedSpeedFocalPlaneExposure(unittest.TestCase):
    """
    Define :func:`colour_hdri.exposure.dsc.\
saturation_based_speed_focal_plane_exposure` definition unit tests methods.
    """

    def test_saturation_based_speed_focal_plane_exposure(self):
        """
        Test :func:`colour_hdri.exposure.dsc.\
saturation_based_speed_focal_plane_exposure` definition.
        """

        np.testing.assert_almost_equal(
            saturation_based_speed_focal_plane_exposure(
                np.array([4000, 2000, 1000]),
                np.array([8, 5.6, 2.8]),
                np.array([1 / 250, 1 / 500, 1 / 1000]),
                np.array([100, 800, 1600]),
                np.array([50 / 1000, 100 / 1000, 500 / 1000]),
                np.array([50 / 1000, 100 / 1000, 500 / 1000]),
                np.array([0.0015, 0.0050, 0.0100]),
                9 / 10,
                98 / 100,
                10,
            ),
            np.array([0.21076116, 0.90368241, 1.90992892]),
            decimal=7,
        )


class TestExposureIndexValues(unittest.TestCase):
    """
    Define :func:`colour_hdri.exposure.dsc.exposure_index_values` definition
    unit tests methods.
    """

    def test_exposure_index_values(self):
        """
        Test :func:`colour_hdri.exposure.dsc.exposure_index_values`
        definition.
        """

        np.testing.assert_almost_equal(
            exposure_index_values(
                np.array([0.16439371, 0.08810904, 0.09310904])
            ),
            np.array([60.82957797, 113.49573211, 107.40095699]),
            decimal=7,
        )


class TestExposureValue100(unittest.TestCase):
    """
    Define :func:`colour_hdri.exposure.dsc.exposure_value_100` definition
    unit tests methods.
    """

    def test_exposure_value_100(self):
        """
        Test :func:`colour_hdri.exposure.dsc.exposure_value_100`
        definition.
        """

        np.testing.assert_almost_equal(
            exposure_value_100(
                np.array([8, 5.6, 2.8]),
                np.array([1 / 250, 1 / 500, 1 / 1000]),
                np.array([100, 800, 1600]),
            ),
            np.array([13.96578428, 10.93663794, 8.93663794]),
            decimal=7,
        )


class TestPhotometricExposureScaleFactorLagarde2014(unittest.TestCase):
    """
    Define :func:`colour_hdri.exposure.dsc.\
photometric_exposure_scale_factor_Lagarde2014` definition unit tests
    methods.
    """

    def test_photometric_exposure_scale_factor_Lagarde2014(self):
        """
        Test :func:`colour_hdri.exposure.dsc.\
photometric_exposure_scale_factor_Lagarde2014` definition.
        """

        np.testing.assert_almost_equal(
            photometric_exposure_scale_factor_Lagarde2014(
                np.array([13.96578428, 10.93663794, 8.93663794]),
                np.array([9 / 10, 8 / 10, 9 / 10]),
                np.array([98 / 100, 98 / 100, 94 / 100]),
                np.array([10, 10, 20]),
            ),
            np.array([0.00005221, 0.00037884, 0.00135554]),
            decimal=7,
        )


if __name__ == "__main__":
    unittest.main()



"""
Adobe DNG SDK Dataset
=====================

Defines various datasets objects for *Adobe DNG SDK*:

-   :attr:`colour_hdri.models.datasets.dng.CCS_ILLUMINANT_ADOBEDNG`
-   :attr:`colour_hdri.models.datasets.dng.\
CCT_ILLUMINANTS_ADOBEDNG`
-   :attr:`colour_hdri.models.datasets.dng.\
LIGHT_SOURCE_TAG_TO_DNG_ILLUMINANTS`

References
----------
-   :cite:`AdobeSystems2015c` : Adobe Systems. (2015). Adobe DNG SDK 1.4 -
    dng_sdk_1_4/dng_sdk/source/dng_camera_profile.cpp -
    dng_camera_profile::IlluminantToTemperature.
    http://download.adobe.com/pub/adobe/dng/dng_sdk_1_4.zip
-   :cite:`AdobeSystems2015d` : Adobe Systems. (2015). Adobe DNG SDK 1.4.
    http://download.adobe.com/pub/adobe/dng/dng_sdk_1_4.zip
-   :cite:`AdobeSystems2015e` : Adobe Systems. (2015). Adobe DNG SDK 1.4 -
    dng_sdk_1_4/dng_sdk/source/dng_tag_values.h - LightSource tag.
    http://download.adobe.com/pub/adobe/dng/dng_sdk_1_4.zip
"""

from __future__ import annotations

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import Dict, NDArray
from colour.utilities import CaseInsensitiveMapping

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CCS_ILLUMINANT_ADOBEDNG",
    "CCT_ILLUMINANTS_ADOBEDNG",
    "LIGHT_SOURCE_TAG_TO_DNG_ILLUMINANTS",
]

CCS_ILLUMINANT_ADOBEDNG: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
]["D50"]
"""*Adobe DNG SDK* default illuminant for *CIE XYZ* colourspace."""

CCT_ILLUMINANTS_ADOBEDNG: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Standard light A": 2850,
        "Tungsten (incandescent light)": 2850,
        "ISO studio tungsten": 3200,
        "D50": 5000,
        "D55": 5500,
        "Daylight": 5500,
        "Fine weather": 5500,
        "Flash": 5500,
        "Standard light B": 5500,
        "D65": 6500,
        "Standard light C": 6500,
        "Cloudy weather": 6500,
        "D75": 7500,
        "Shade": 7500,
        "Daylight fluorescent (D 5700 - 7100K)": (5700 + 7100) * 0.5,
        "Day white fluorescent (N 4600 - 5500K)": (4600 + 5400) * 0.5,
        "Cool white fluorescent (W 3800 - 4500K)": (3900 + 4500) * 0.5,
        "Fluorescent": (3900 + 4500) * 0.5,
        "White fluorescent (WW 3250 - 3800K)": (3200 + 3700) * 0.5,
        "Warm white fluorescent (L 2600 - 3250K)": (2600 + 3250) * 0.5,
    }
)
"""
*Adobe DNG SDK* illuminants correlated colour temperature.

References
----------
:cite:`AdobeSystems2015c`

Notes
-----
-   The correlated colour temperature are given for the
    *CIE 1931 2 Degree Standard Observer*.
"""

LIGHT_SOURCE_TAG_TO_DNG_ILLUMINANTS: Dict = {
    1: "Daylight",
    2: "Fluorescent",
    3: "Tungsten (incandescent light)",
    4: "Flash",
    9: "Fine weather",
    10: "Cloudy weather",
    11: "Shade",
    12: "Daylight fluorescent (D 5700 - 7100K)",
    13: "Day white fluorescent (N 4600 - 5500K)",
    14: "Cool white fluorescent (W 3800 - 4500K)",
    15: "White fluorescent (WW 3250 - 3800K)",
    16: "Warm white fluorescent (L 2600 - 3250K)",
    17: "Standard light A",
    18: "Standard light B",
    19: "Standard light C",
    20: "D55",
    21: "D65",
    22: "D75",
    23: "D50",
    24: "ISO studio tungsten",
    255: "Other",
}
"""
*Adobe DNG SDK* *LightSource Tag* indexes mapping to illuminants.

References
----------
:cite:`AdobeSystems2015e`
"""



"""
Viriyothai (2009) Variance Minimization Light Probe Sampling
============================================================

Defines the *Viriyothai (2009)* variance minimization light probe sampling
objects:

-   :func:`colour_hdri.\
light_probe_sampling_variance_minimization_Viriyothai2009`

References
----------
-   :cite:`Viriyothai2009` : Viriyothai, K., & Debevec, P. (2009). Variance
    minimization light probe sampling. SIGGRAPH "09: Posters on - SIGGRAPH "09,
    Egsr, 1-1. doi:10.1145/1599301.1599393
"""

import numpy as np
from collections import namedtuple

from colour.models import RGB_COLOURSPACES, RGB_luminance
from colour.utilities import as_float_array, centroid, warning

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "Light_Specification",
    "luminance_variance",
    "find_regions_variance_minimization_Viriyothai2009",
    "highlight_regions_variance_minimization",
    "light_probe_sampling_variance_minimization_Viriyothai2009",
]


class Light_Specification(
    namedtuple("Light_Specification", ("uv", "colour", "index"))
):
    """
    Define a light probe sampling resulting light specification.

    Parameters
    ----------
    uv : array_like
        :math:`uv` coordinates of the sampled light.
    colour : array_like
        Sampled light colour.
    index : array_like
        Sampled light location index in its original array.
    """


def luminance_variance(a):
    """
    Compute the Luminance variance of given :math:`a` 2-D array.

    Parameters
    ----------
    a : array_like
        :math:`a` 2-D array to compute the Luminance variation.

    Returns
    -------
    numeric
        :math:`a` 2-D array Luminance variance.

    Examples
    --------
    >>> a = np.tile(np.arange(5), (5, 1))
    >>> luminance_variance(a)  # doctest: +ELLIPSIS
    12.2474487...
    """

    x, y = np.mgrid[0 : np.shape(a)[0], 0 : np.shape(a)[1]]

    x_centroid, y_centroid = centroid(a)

    variance = np.sqrt(
        np.sum(a * ((y - y_centroid) ** 2 + (x - x_centroid) ** 2))
    )

    return variance


def find_regions_variance_minimization_Viriyothai2009(a, n=4):
    """
    Find the :math:`2^n` regions using *Viriyothai (2009)* variance
    minimization light probe sampling algorithm on given :math:`a` 2-D array.

    Parameters
    ----------
    a : array_like
        :math:`a` 2-D array to find the regions.
    n : int
        Iterations count, the total regions count is :math:`2^n`.

    Returns
    -------
    list
        Regions with variance minimized.
    """

    a = as_float_array(a)

    regions = [(0, a.shape[0], 0, a.shape[1])]

    for _i in range(n):
        sub_regions = []
        for region in regions:
            variance = np.inf
            horizontal = False
            location = -1
            y_min, y_max, x_min, x_max = region

            # Vertical cut on image width.
            for j in range(x_min, x_max):
                variance_c = max(
                    luminance_variance(a[y_min:y_max, x_min:j]),
                    luminance_variance(a[y_min:y_max, j:x_max]),
                )

                if variance_c < variance:
                    variance = variance_c
                    horizontal = False
                    location = j

            # Horizontal cut on image height.
            for j in range(y_min, y_max):
                variance_c = max(
                    luminance_variance(a[y_min:j, x_min:x_max]),
                    luminance_variance(a[j:y_max, x_min:x_max]),
                )

                if variance_c < variance:
                    variance = variance_c
                    horizontal = True
                    location = j

            if horizontal:
                sub_regions.append((y_min, location, x_min, x_max))
                sub_regions.append((location, y_max, x_min, x_max))
            else:
                sub_regions.append((y_min, y_max, x_min, location))
                sub_regions.append((y_min, y_max, location, x_max))

        regions = sub_regions

    return regions


def highlight_regions_variance_minimization(
    a, regions, highlight_colour=np.array([0, 1, 0])
):
    """
    Highlight regions using with variance minimized on given :math:`a`
    3-D array.

    Parameters
    ----------
    a : array_like
        :math:`a` 3-D array to highlight the regions.
    regions : array_like
        Regions with variance minimized.
    highlight_colour : array_like
        Highlight colour.

    Returns
    -------
    ndarray
        :math:`a` 3-D array with highlighted regions.
    """

    a = np.copy(a)

    for region in regions:
        y_min, y_max, x_min, x_max = region

        a[y_min:y_max, x_min] = highlight_colour
        a[y_min:y_max, x_max - 1] = highlight_colour
        a[y_min, x_min:x_max] = highlight_colour
        a[y_max - 1, x_min:x_max] = highlight_colour

    return a


def light_probe_sampling_variance_minimization_Viriyothai2009(
    light_probe, lights_count=16, colourspace=RGB_COLOURSPACES["sRGB"]
):
    """
    Sample given light probe to find lights using *Viriyothai (2009)* variance
    minimization light probe sampling algorithm.

    Parameters
    ----------
    light_probe : array_like
        Array to sample for lights.
    lights_count : int
        Amount of lights to generate.
    colourspace : `colour.RGB_Colourspace`, optional
        *RGB* colourspace used for internal *Luminance* computation.

    Returns
    -------
    list
        list of
        :class:`colour_hdri.sampling.variance_minimization.Light_Specification`
        lights.

    References
    ----------
    :cite:`Viriyothai2009`
    """

    light_probe = as_float_array(light_probe)

    iterations = np.sqrt(lights_count).astype(np.int_)
    if iterations**2 != lights_count:
        warning(
            f"{lights_count} lights requested, {iterations ** 2} will be "
            f"effectively computed!"
        )

    Y = RGB_luminance(
        light_probe, colourspace.primaries, colourspace.whitepoint
    )
    regions = find_regions_variance_minimization_Viriyothai2009(Y, iterations)

    lights = []
    for region in regions:
        y_min, y_max, x_min, x_max = region
        c = centroid(Y[y_min:y_max, x_min:x_max])
        c = c + np.array([y_min, x_min])
        light_probe_c = light_probe[y_min:y_max, x_min:x_max]
        lights.append(
            Light_Specification(
                (c / np.array(Y.shape))[::-1],
                np.sum(np.sum(light_probe_c, 0), 0),
                c,
            )
        )

    return lights



"""
HDRI / Radiance Image Generation
================================

Defines the HDRI / radiance image generation objects:

-   :func:`colour_hdri.image_stack_to_radiance_image`

See Also
--------
`Colour - HDRI - Examples Jupyter Notebooks
<https://github.com/colour-science/colour-hdri/\
blob/master/colour_hdri/examples>`__

References
----------
-   :cite:`Banterle2011n` : Banterle, F., Artusi, A., Debattista, K., &
    Chalmers, A. (2011). 2.1.1 Generating HDR Content by Combining Multiple
    Exposures. In Advanced High Dynamic Range Imaging. A K Peters/CRC Press.
    ISBN:978-1-56881-719-4
"""

from __future__ import annotations

import numpy as np

from colour.utilities import as_float_array, tsplit, tstack, warning
from colour.hints import ArrayLike, Boolean, Callable, NDArray, Optional

from colour_hdri.exposure import average_luminance
from colour_hdri.generation import weighting_function_Debevec1997
from colour_hdri.utilities import ImageStack

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "image_stack_to_radiance_image",
]


def image_stack_to_radiance_image(
    image_stack: ImageStack,
    weighting_function: Callable = weighting_function_Debevec1997,
    weighting_average: Boolean = False,
    camera_response_functions: Optional[ArrayLike] = None,
) -> Optional[NDArray]:
    """
    Generate a HDRI / radiance image from given image stack.

    Parameters
    ----------
    image_stack
        Stack of single channel or multi-channel floating point images. The
        stack is assumed to be representing linear values except if
        ``camera_response_functions`` argument is provided.
    weighting_function
        Weighting function :math:`w`.
    weighting_average
         Enables weighting function :math:`w` computation on channels average
         instead of on a per-channel basis.
    camera_response_functions
        Camera response functions :math:`g(z)` of the imaging system / camera
        if the stack is representing non-linear values.

    Returns
    -------
    :class:`numpy.ndarray`
        Radiance image.

    Warnings
    --------
    If the image stack contains images with negative or equal to zero values,
    unpredictable results may occur and NaNs might be generated. It is
    thus recommended encoding the images in a wider RGB colourspace or clamp
    negative values.

    References
    ----------
    :cite:`Banterle2011n`
    """

    image_c: Optional[NDArray] = None
    weight_c: Optional[NDArray] = None
    for i, image in enumerate(image_stack):
        if image.data is not None and image.metadata is not None:
            if image_c is None:
                image_c = np.zeros(image.data.shape)
                weight_c = np.zeros(image.data.shape)

            L = 1 / average_luminance(
                image.metadata.f_number,
                image.metadata.exposure_time,
                image.metadata.iso,
            )

            if np.any(image.data <= 0):
                warning(
                    f'"{image.path}" image channels contain negative or equal '
                    f"to zero values, unpredictable results may occur! Please "
                    f"consider encoding your images in a wider gamut RGB "
                    f"colourspace or clamp negative values."
                )

            if weighting_average and image.data.ndim == 3:
                average = np.average(image.data, axis=-1)

                weights = weighting_function(average)
                weights = np.rollaxis(weights[np.newaxis], 0, 3)
                if i == 0:
                    weights[average >= 0.5] = 1
                if i == len(image_stack) - 1:
                    weights[average <= 0.5] = 1
            else:
                weights = weighting_function(image.data)
                if i == 0:
                    weights[image.data >= 0.5] = 1
                if i == len(image_stack) - 1:
                    weights[image.data <= 0.5] = 1

            image_data = image.data
            if camera_response_functions is not None:
                camera_response_functions = as_float_array(
                    camera_response_functions
                )
                samples = np.linspace(0, 1, camera_response_functions.shape[0])

                R, G, B = tsplit(image.data)
                R = np.interp(R, samples, camera_response_functions[..., 0])
                G = np.interp(G, samples, camera_response_functions[..., 1])
                B = np.interp(B, samples, camera_response_functions[..., 2])
                image_data = tstack([R, G, B])

            image_c += weights * image_data / L
            weight_c += weights

    if image_c is not None and weight_c is not None:
        image_c /= weight_c

    return image_c



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour_hdri.utilities.common` module."""

from __future__ import annotations

import os
import unittest

from colour_hdri import TESTS_RESOURCES_DIRECTORY
from colour_hdri.utilities import (
    vivification,
    vivified_to_dict,
    path_exists,
    filter_files,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "FROBISHER_001_DIRECTORY",
    "TestVivification",
    "TestVivifiedToDict",
    "TestPathExists",
    "TestFilterFiles",
]

FROBISHER_001_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "frobisher_001"
)


class TestVivification(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.common.vivification` definition unit
    tests methods.
    """

    def test_vivification(self):
        """Test :func:`colour_hdri.utilities.common.vivification` definition."""

        vivified = vivification()
        vivified["my"]["attribute"] = 1
        self.assertIn("attribute", vivified["my"].keys())
        self.assertEqual(vivified["my"]["attribute"], 1)


class TestVivifiedToDict(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.common.vivified_to_dict` definition
    unit tests methods.
    """

    def test_vivified_to_dict(self):
        """Test :func:`colour_hdri.utilities.common.vivified_to_dict` definition."""

        vivified = vivification()
        vivified["my"]["attribute"] = 1
        vivified_as_dict = vivified_to_dict(vivified)
        self.assertIsInstance(dict(), type(vivified_as_dict))
        self.assertIn("attribute", vivified_as_dict["my"].keys())
        self.assertEqual(vivified_as_dict["my"]["attribute"], 1)


class TestPathExists(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.common.path_exists` definition unit
    tests methods.
    """

    def test_path_exists(self):
        """Test :func:`colour_hdri.utilities.common.path_exists` definition."""

        self.assertTrue(path_exists(__file__))
        self.assertFalse(path_exists(""))


class TestFilterFiles(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.common.filter_files` definition unit
    tests methods.
    """

    def test_filter_files(self):
        """Test :func:`colour_hdri.utilities.common.filter_files` definition."""

        raw_files = filter_files(FROBISHER_001_DIRECTORY, ("CR2", "jpg"))
        self.assertListEqual(
            sorted(map(os.path.basename, raw_files)),
            [
                "IMG_2600.CR2",
                "IMG_2600.jpg",
                "IMG_2601.CR2",
                "IMG_2601.jpg",
                "IMG_2602.CR2",
                "IMG_2602.jpg",
            ],
        )


if __name__ == "__main__":
    unittest.main()



"""
Tonemapping Operators Plotting
==============================

Defines the tonemapping operators plotting objects:

-   :func:`colour_hdri.plotting.plot_tonemapping_operator_image`
"""

from __future__ import annotations

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.ticker
import numpy as np

from colour.hints import Any, ArrayLike, Boolean, Callable, Dict, Tuple
from colour.plotting import (
    CONSTANTS_COLOUR_STYLE,
    artist,
    override_style,
    render,
)
from colour.utilities import as_float_array

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "plot_tonemapping_operator_image",
]


@override_style()
def plot_tonemapping_operator_image(
    image: ArrayLike,
    luminance_function: ArrayLike,
    log_scale: Boolean = False,
    cctf_encoding: Callable = CONSTANTS_COLOUR_STYLE.colour.colourspace.cctf_encoding,
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot given tonemapped image with superimposed luminance mapping function.

    Parameters
    ----------
    image
         Tonemapped image to plot.
    luminance_function
        Luminance mapping function.
    log_scale
        Use a log scale for plotting the luminance mapping function.
    cctf_encoding
        Encoding colour component transfer function / opto-electronic
        transfer function used for plotting.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.render`},
        Please refer to the documentation of the previously listed definition.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.
    """

    image = as_float_array(image)
    luminance_function = as_float_array(luminance_function)

    settings: Dict[str, Any] = {"uniform": True}
    settings.update(kwargs)

    figure, axes = artist(**settings)

    shape = image.shape
    bounding_box = [0, 1, 0, 1]

    image = np.clip(cctf_encoding(image), 0, 1)
    axes.imshow(
        image,
        aspect=shape[0] / shape[1],
        extent=bounding_box,
        interpolation="nearest",
    )

    axes.plot(
        np.linspace(0, 1, len(luminance_function)),
        luminance_function,
        color="red",
    )

    settings = {
        "axes": axes,
        "bounding_box": bounding_box,
        "x_ticker": True,
        "y_ticker": True,
        "x_label": "Input Luminance",
        "y_label": "Output Luminance",
    }
    settings.update(kwargs)

    if log_scale:
        settings.update(
            {
                "x_label": "$log_2$ Input Luminance",
                "x_ticker_locator": matplotlib.ticker.AutoMinorLocator(0.5),
            }
        )
        plt.gca().set_xscale("log", basex=2)
        plt.gca().xaxis.set_major_formatter(
            matplotlib.ticker.ScalarFormatter()
        )

    return render(**settings)



"""
Absolute Luminance Calibration - Lagarde (2016)
===============================================

Defines the *Lagarde (2016)* panoramic images absolute *Luminance* calibration
objects:

-   :func:`colour_hdri.absolute_luminance_calibration_Lagarde2016`
-   :func:`colour_hdri.upper_hemisphere_illuminance_Lagarde2016`
-   :func:`colour_hdri.upper_hemisphere_illuminance_weights_Lagarde2016`

References
----------
-   :cite:`Lagarde2016b` : Lagarde, Sebastien, Lachambre, S., & Jover, C.
    (2016). An Artist-Friendly Workflow for Panoramic HDRI.
    http://blog.selfshadow.com/publications/s2016-shading-course/unity/\
s2016_pbs_unity_hdri_notes.pdf
"""

from __future__ import annotations

import numpy as np

from colour import RGB_COLOURSPACES, RGB_Colourspace, RGB_luminance
from colour.hints import ArrayLike, Floating, Integer, NDArray
from colour.utilities import as_float_array

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "upper_hemisphere_illuminance_Lagarde2016",
    "upper_hemisphere_illuminance_weights_Lagarde2016",
    "absolute_luminance_calibration_Lagarde2016",
]


def upper_hemisphere_illuminance_Lagarde2016(
    RGB: ArrayLike, colourspace: RGB_Colourspace = RGB_COLOURSPACES["sRGB"]
) -> Floating:
    """
    Compute upper hemisphere illuminance :math:`E_v` of given RGB panoramic
    image.

    Parameters
    ----------
    RGB
        *RGB* panoramic image array.
    colourspace
        *RGB* colourspace used for internal *Luminance* computation.

    Returns
    -------
    :class:`numpy.floating`
        Upper hemisphere illuminance :math:`E_v`.

    References
    ----------
    :cite:`Lagarde2016b`

    Examples
    --------
    >>> RGB = np.ones((16, 32, 3))
    >>> upper_hemisphere_illuminance_Lagarde2016(RGB)  # doctest: +ELLIPSIS
    2.9344691...
    """

    RGB = as_float_array(RGB)

    height, width, _channels = RGB.shape

    L = RGB_luminance(RGB, colourspace.primaries, colourspace.whitepoint)

    theta = np.linspace(0, 1, height) * np.pi

    theta_cos = np.cos(theta)[..., np.newaxis]
    theta_sin = np.sin(theta)[..., np.newaxis]

    E_v = np.sum(np.where(theta_cos > 0, L * theta_cos * theta_sin, 0))

    E_v *= 2 * np.pi**2 / (width * height)

    return E_v


def upper_hemisphere_illuminance_weights_Lagarde2016(
    height: Integer, width: Integer
) -> NDArray:
    """
    Compute upper hemisphere illuminance weights for use with applications
    unable to perform the computation directly, i.e. *Adobe Photoshop*.

    Parameters
    ----------
    height
        Output array height.
    width
        Output array width.

    Returns
    -------
    :class:`numpy.ndarray`
        Upper hemisphere illuminance weights.

    References
    ----------
    :cite:`Lagarde2016b`

    Examples
    --------
    >>> upper_hemisphere_illuminance_weights_Lagarde2016(  # doctest: +ELLIPSIS
    ...    16, 1)
    array([[ 0...        ],
           [ 4.0143297...],
           [ 7.3345454...],
           [ 9.3865515...],
           [ 9.8155376...],
           [ 8.5473281...],
           [ 5.8012079...],
           [ 2.0520061...],
           [ 0...        ],
           [ 0...        ],
           [ 0...        ],
           [ 0...        ],
           [ 0...        ],
           [ 0...        ],
           [ 0...        ],
           [ 0...        ]])
    """

    w = np.zeros((height, width))

    theta = np.linspace(0, 1, height) * np.pi
    theta = np.tile(theta[..., np.newaxis], (1, width))

    theta_cos = np.cos(theta)
    theta_sin = np.sin(theta)

    w[theta_cos > 0] = (
        theta_cos[theta_cos > 0] * theta_sin[theta_cos > 0] * 2 * np.pi**2
    )

    return w


def absolute_luminance_calibration_Lagarde2016(
    RGB: ArrayLike,
    measured_illuminance: Floating,
    colourspace: RGB_Colourspace = RGB_COLOURSPACES["sRGB"],
) -> NDArray:
    """
    Perform absolute *Luminance* calibration of given *RGB* panoramic image
    using *Lagarde (2016)* method.

    Parameters
    ----------
    RGB
        *RGB* panoramic image to calibrate.
    measured_illuminance
        Measured illuminance :math:`E_v`.
    colourspace
        *RGB* colourspace used for internal *Luminance* computation.

    Returns
    -------
    :class:`numpy.ndarray`
         Absolute *Luminance* calibrated *RGB* panoramic image.

    Examples
    --------
    >>> RGB = np.ones((4, 8, 3))
    >>> absolute_luminance_calibration_Lagarde2016(  # doctest: +ELLIPSIS
    ...     RGB, 500)
    array([[[ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...]],
    <BLANKLINE>
           [[ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...]],
    <BLANKLINE>
           [[ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...]],
    <BLANKLINE>
           [[ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...],
            [ 233.9912506...,  233.9912506...,  233.9912506...]]])
    """

    RGB = as_float_array(RGB)

    E_v = upper_hemisphere_illuminance_Lagarde2016(RGB, colourspace)

    return RGB / E_v * measured_illuminance



# !/usr/bin/env python
"""
Defines the unit tests for the
:mod:`colour_hdri.generation.weighting_functions` module.
"""

import numpy as np
import unittest

from colour_hdri.generation import (
    normal_distribution_function,
    hat_function,
    weighting_function_Debevec1997,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestNormalDistributionFunction",
    "TestHatFunction",
    "TestWeightingFunctionDebevec1997",
]


class TestNormalDistributionFunction(unittest.TestCase):
    """
    Define :func:`colour_hdri.generation.weighting_functions.\
normal_distribution_function` definition unit tests methods.
    """

    def test_normal_distribution_function(self):
        """
        Test :func:`colour_hdri.generation.weighting_functions.\
normal_distribution_function` definition.
        """

        np.testing.assert_almost_equal(
            normal_distribution_function(np.linspace(0, 1, 10)),
            np.array(
                [
                    0.00386592,
                    0.03470859,
                    0.18002174,
                    0.53940751,
                    0.93371212,
                    0.93371212,
                    0.53940751,
                    0.18002174,
                    0.03470859,
                    0.00386592,
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            normal_distribution_function(np.linspace(0, 1, 10), 0, 1),
            np.array(
                [
                    1.00000000,
                    0.99384617,
                    0.97561098,
                    0.94595947,
                    0.90595519,
                    0.85699689,
                    0.80073740,
                    0.73899130,
                    0.67363846,
                    0.60653066,
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            normal_distribution_function(np.linspace(0, 1, 10), 0.5, 0.5),
            np.array(
                [
                    0.60653066,
                    0.73899130,
                    0.85699689,
                    0.94595947,
                    0.99384617,
                    0.99384617,
                    0.94595947,
                    0.85699689,
                    0.73899130,
                    0.60653066,
                ]
            ),
            decimal=7,
        )


class TestHatFunction(unittest.TestCase):
    """
    Define :func:`colour_hdri.generation.weighting_functions.hat_function`
    definition unit tests methods.
    """

    def test_hat_function(self):
        """
        Test :func:`colour_hdri.generation.weighting_functions.hat_function`
        definition.
        """

        np.testing.assert_almost_equal(
            hat_function(np.linspace(0, 1, 10)),
            np.array(
                [
                    0.00000000,
                    0.95099207,
                    0.99913557,
                    0.99999812,
                    1.00000000,
                    1.00000000,
                    0.99999812,
                    0.99913557,
                    0.95099207,
                    0.00000000,
                ]
            ),
            decimal=7,
        )


class TestWeightingFunctionDebevec1997(unittest.TestCase):
    """
    Define :func:`colour_hdri.generation.weighting_functions.\
weighting_function_Debevec1997` definition unit tests methods.
    """

    def test_weighting_function_Debevec1997(self):
        """
        Test :func:`colour_hdri.generation.weighting_functions.\
weighting_function_Debevec1997` definition.
        """

        np.testing.assert_almost_equal(
            weighting_function_Debevec1997(np.linspace(0, 1, 10)),
            np.array(
                [
                    0.00000000,
                    0.23273657,
                    0.48849105,
                    0.74424552,
                    1.00000000,
                    1.00000000,
                    0.74424552,
                    0.48849105,
                    0.23273657,
                    0.00000000,
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            weighting_function_Debevec1997(np.linspace(0, 1, 10), 0, 1),
            np.array(
                [
                    0.00000000,
                    0.25000000,
                    0.50000000,
                    0.75000000,
                    1.00000000,
                    1.00000000,
                    0.75000000,
                    0.50000000,
                    0.25000000,
                    0.00000000,
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            weighting_function_Debevec1997(np.linspace(0, 1, 10), 0.25, 0.75),
            np.array(
                [
                    0.00000000,
                    0.00000000,
                    0.00000000,
                    0.42857143,
                    1.00000000,
                    1.00000000,
                    0.42857143,
                    0.00000000,
                    0.00000000,
                    0.00000000,
                ]
            ),
            decimal=7,
        )


if __name__ == "__main__":
    unittest.main()



from .dng import (
    RAW_CONVERTER,
    RAW_CONVERSION_ARGUMENTS,
    RAW_D_CONVERSION_ARGUMENTS,
    DNG_CONVERTER,
    DNG_CONVERSION_ARGUMENTS,
    DNG_EXIF_TAGS_BINDING,
    convert_raw_files_to_dng_files,
    convert_dng_files_to_intermediate_files,
    read_dng_files_exif_tags,
)

__all__ = [
    "RAW_CONVERTER",
    "RAW_CONVERSION_ARGUMENTS",
    "RAW_D_CONVERSION_ARGUMENTS",
    "DNG_CONVERTER",
    "DNG_CONVERSION_ARGUMENTS",
    "DNG_EXIF_TAGS_BINDING",
    "convert_raw_files_to_dng_files",
    "convert_dng_files_to_intermediate_files",
    "read_dng_files_exif_tags",
]



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour_hdri.utilities.exif` module."""

from __future__ import annotations

import numpy as np
import os
import shutil
import tempfile
import unittest

from colour_hdri import TESTS_RESOURCES_DIRECTORY
from colour_hdri.utilities import filter_files, vivified_to_dict
from colour_hdri.utilities import (
    EXIFTag,
    parse_exif_string,
    parse_exif_number,
    parse_exif_fraction,
    parse_exif_array,
    parse_exif_data,
    read_exif_tags,
    copy_exif_tags,
    update_exif_tags,
    delete_exif_tags,
    read_exif_tag,
    write_exif_tag,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "FROBISHER_001_DIRECTORY",
    "TestParseExifString",
    "TestParseExifNumber",
    "TestParseExifFraction",
    "TestParseExifArray",
    "TestParseExifData",
    "TestReadExifTags",
    "TestCopyExifTags",
    "TestUpdateExifTags",
    "TestDeleteExifTags",
    "TestReadExifTag",
    "TestWriteExifTag",
]

FROBISHER_001_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "frobisher_001"
)


class TestParseExifString(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.exif.parse_exif_string` definition
    unit tests methods.
    """

    def test_parse_exif_string(self):
        """Test :func:`colour_hdri.utilities.exif.parse_exif_string` definition."""

        exif_tag = EXIFTag("EXIF", "Make", "Canon", "271")
        self.assertEqual(parse_exif_string(exif_tag), "Canon")


class TestParseExifNumber(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.exif.parse_exif_number` definition
    unit tests methods.
    """

    def test_parse_exif_number(self):
        """Test :func:`colour_hdri.utilities.exif.parse_exif_number` definition."""

        exif_tag = EXIFTag("EXIF", "Focal Length", "16", "37386")
        self.assertEqual(parse_exif_number(exif_tag), 16)

        exif_tag = EXIFTag("EXIF", "Focal Length", "16", "37386")
        self.assertIsInstance(parse_exif_number(exif_tag, np.int_), np.int_)


class TestParseExifFraction(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.exif.parse_exif_fraction` definition
    unit tests methods.
    """

    def test_parse_exif_fraction(self):
        """
        Test :func:`colour_hdri.utilities.exif.parse_exif_fraction`
        definition.
        """

        exif_tag = EXIFTag("EXIF", "Exposure Time", "0.01666666667", "33434")
        self.assertAlmostEqual(
            parse_exif_fraction(exif_tag), 0.01666666, places=7
        )

        exif_tag = EXIFTag("EXIF", "Exposure Time", "10/4000", "33434")
        self.assertAlmostEqual(
            parse_exif_fraction(exif_tag), 0.00250000, places=7
        )


class TestParseExifArray(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.exif.parse_exif_array` definition
    unit tests methods.
    """

    def test_parse_exif_array(self):
        """Test :func:`colour_hdri.utilities.exif.parse_exif_array` definition."""

        exif_tag = EXIFTag(
            "EXIF",
            "Color Matrix 1",
            (
                "0.5309 -0.0229 -0.0336 "
                "-0.6241 1.3265 0.3337 "
                "-0.0817 0.1215 0.6664"
            ),
            "50721",
        )
        np.testing.assert_array_equal(
            parse_exif_array(exif_tag),
            np.array(
                [
                    0.5309,
                    -0.0229,
                    -0.0336,
                    -0.6241,
                    1.3265,
                    0.3337,
                    -0.0817,
                    0.1215,
                    0.6664,
                ]
            ),
        )

        np.testing.assert_array_equal(
            parse_exif_array(exif_tag, shape=(3, 3)),
            np.array(
                [
                    [0.5309, -0.0229, -0.0336],
                    [-0.6241, 1.3265, 0.3337],
                    [-0.0817, 0.1215, 0.6664],
                ]
            ),
        )


class TestParseExifData(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.exif.parse_exif_data` definition
    unit tests methods.
    """

    def test_parse_exif_data(self):
        """Test :func:`colour_hdri.utilities.exif.parse_exif_data` definition."""

        self.assertListEqual(
            parse_exif_data(
                "[XMP]               - Description                     :"
            ),
            ["XMP", "-", "Description", ""],
        )

        self.assertListEqual(
            parse_exif_data(
                "[EXIF]            296 Resolution Unit                 : 2"
            ),
            ["EXIF", "296", "Resolution Unit", "2"],
        )

        self.assertListEqual(
            parse_exif_data(
                "[ICC_Profile]       8 Profile Version                 : 528"
            ),
            ["ICC_Profile", "8", "Profile Version", "528"],
        )


class TestReadExifTags(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.exif.read_exif_tags` definition unit
    tests methods.
    """

    def test_read_exif_tags(self):
        """Test :func:`colour_hdri.utilities.exif.read_exif_tags` definition."""

        test_jpg_image = filter_files(FROBISHER_001_DIRECTORY, ("jpg",))[0]
        exif_data = vivified_to_dict(read_exif_tags(test_jpg_image))

        self.assertIsInstance(exif_data, type(dict()))

        self.assertListEqual(
            sorted(exif_data.keys()),
            [
                "Composite",
                "EXIF",
                "ExifTool",
                "File",
                "ICC_Profile",
                "JFIF",
                "Photoshop",
                "XMP",
            ],
        )

        self.assertListEqual(
            sorted(exif_data["EXIF"].values(), key=lambda x: x[0].name),
            [
                [
                    EXIFTag(
                        "EXIF", "Camera Model Name", "EOS 5D Mark II", "272"
                    )
                ],
                [
                    EXIFTag(
                        "EXIF", "Create Date", "2015:09:19 03:39:20", "36868"
                    )
                ],
                [
                    EXIFTag(
                        "EXIF",
                        "Date/Time Original",
                        "2015:09:19 03:39:20",
                        "36867",
                    )
                ],
                [EXIFTag("EXIF", "Exif Image Height", "426", "40963")],
                [EXIFTag("EXIF", "Exif Image Width", "640", "40962")],
                [EXIFTag("EXIF", "Exposure Time", "0.125", "33434")],
                [EXIFTag("EXIF", "F Number", "8", "33437")],
                [EXIFTag("EXIF", "Focal Length", "16", "37386")],
                [EXIFTag("EXIF", "ISO", "100", "34855")],
                [EXIFTag("EXIF", "Make", "Canon", "271")],
                [EXIFTag("EXIF", "Modify Date", "2015:09:19 03:39:20", "306")],
                [EXIFTag("EXIF", "Orientation", "1", "274")],
                [EXIFTag("EXIF", "Photometric Interpretation", "2", "262")],
                [EXIFTag("EXIF", "Resolution Unit", "2", "296")],
                [EXIFTag("EXIF", "Software", "Photos 1.0.1", "305")],
                [EXIFTag("EXIF", "X Resolution", "72", "282")],
                [EXIFTag("EXIF", "Y Resolution", "72", "283")],
            ],
        )


class TestCopyExifTags(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.exif.copy_exif_tags` definition unit
    tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._temporary_directory = tempfile.mkdtemp()

    def tearDown(self):
        """After tests actions."""

        shutil.rmtree(self._temporary_directory)

    def test_copy_exif_tags(self):
        """Test :func:`colour_hdri.utilities.exif.copy_exif_tags` definition."""

        reference_jpg_image = filter_files(FROBISHER_001_DIRECTORY, ("jpg",))[
            0
        ]
        test_jpg_image = os.path.join(
            self._temporary_directory, os.path.basename(reference_jpg_image)
        )

        shutil.copyfile(reference_jpg_image, test_jpg_image)
        self.assertEqual(read_exif_tag(test_jpg_image, "Aperture"), "8.0")
        delete_exif_tags(test_jpg_image)
        self.assertEqual(read_exif_tag(test_jpg_image, "Aperture"), "")
        copy_exif_tags(reference_jpg_image, test_jpg_image)
        self.assertEqual(read_exif_tag(test_jpg_image, "Aperture"), "8.0")


class TestUpdateExifTags(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.exif.update_exif_tags` definition unit
    tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._temporary_directory = tempfile.mkdtemp()

    def tearDown(self):
        """After tests actions."""

        shutil.rmtree(self._temporary_directory)

    def test_update_exif_tags(self):
        """Test :func:`colour_hdri.utilities.exif.update_exif_tags` definition."""

        reference_jpg_images = filter_files(FROBISHER_001_DIRECTORY, ("jpg",))
        test_jpg_images = []
        for reference_jpg_image in reference_jpg_images:
            test_jpg_image = os.path.join(
                self._temporary_directory,
                os.path.basename(reference_jpg_image),
            )
            shutil.copyfile(reference_jpg_image, test_jpg_image)
            delete_exif_tags(test_jpg_image)
            self.assertEqual(read_exif_tag(test_jpg_image, "Aperture"), "")
            test_jpg_images.append(test_jpg_image)

        update_exif_tags(zip(reference_jpg_images, test_jpg_images))
        for test_jpg_image in test_jpg_images:
            self.assertEqual(read_exif_tag(test_jpg_image, "Aperture"), "8.0")


class TestDeleteExifTags(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.exif.delete_exif_tags` definition unit
    tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._temporary_directory = tempfile.mkdtemp()

    def tearDown(self):
        """After tests actions."""

        shutil.rmtree(self._temporary_directory)

    def test_delete_exif_tags(self):
        """Test :func:`colour_hdri.utilities.exif.delete_exif_tags` definition."""

        reference_jpg_image = filter_files(FROBISHER_001_DIRECTORY, ("jpg",))[
            0
        ]
        test_jpg_image = os.path.join(
            self._temporary_directory, os.path.basename(reference_jpg_image)
        )

        shutil.copyfile(reference_jpg_image, test_jpg_image)
        self.assertEqual(read_exif_tag(test_jpg_image, "Aperture"), "8.0")
        delete_exif_tags(test_jpg_image)
        self.assertEqual(read_exif_tag(test_jpg_image, "Aperture"), "")


class TestReadExifTag(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.exif.read_exif_tag` definition unit
    tests methods.
    """

    def test_read_exif_tag(self):
        """Test :func:`colour_hdri.utilities.exif.read_exif_tag` definition."""

        test_jpg_image = filter_files(FROBISHER_001_DIRECTORY, ("jpg",))[0]

        self.assertEqual(read_exif_tag(test_jpg_image, "Aperture"), "8.0")
        self.assertEqual(read_exif_tag(test_jpg_image, "ExposureTime"), "1/8")
        self.assertEqual(read_exif_tag(test_jpg_image, "ISO"), "100")


class TestWriteExifTag(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.exif.write_exif_tag` definition unit
    tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._temporary_directory = tempfile.mkdtemp()

    def tearDown(self):
        """After tests actions."""

        shutil.rmtree(self._temporary_directory)

    def test_write_exif_tag(self):
        """Test :func:`colour_hdri.utilities.exif.write_exif_tag` definition."""

        reference_jpg_image = filter_files(FROBISHER_001_DIRECTORY, ("jpg",))[
            0
        ]
        test_jpg_image = os.path.join(
            self._temporary_directory, os.path.basename(reference_jpg_image)
        )

        shutil.copyfile(reference_jpg_image, test_jpg_image)
        # *Aperture* exif tag is not writeable, changing for *FNumber*.
        self.assertEqual(read_exif_tag(test_jpg_image, "FNumber"), "8.0")
        write_exif_tag(test_jpg_image, "FNumber", "16.0")
        self.assertEqual(read_exif_tag(test_jpg_image, "FNumber"), "16.0")


if __name__ == "__main__":
    unittest.main()



# !/usr/bin/env python
"""
Defines the unit tests for the
:mod:`colour_hdri.calibration.absolute_luminance` module.
"""

from __future__ import annotations

import numpy as np
import os
import unittest

from colour import read_image

from colour_hdri import TESTS_RESOURCES_DIRECTORY
from colour_hdri.calibration import (
    upper_hemisphere_illuminance_weights_Lagarde2016,
    absolute_luminance_calibration_Lagarde2016,
)
from colour_hdri.calibration.absolute_luminance import (
    upper_hemisphere_illuminance_Lagarde2016,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestUpperHemisphereIlluminanceLagarde2016",
    "TestUpperHemisphereIlluminanceWeightsLagarde2016",
    "TestAbsoluteLuminanceCalibrationLagarde2016",
]

UNITY_001_DIRECTORY: str = os.path.join(TESTS_RESOURCES_DIRECTORY, "unity_001")

CALIBRATION_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "colour_hdri", "calibration"
)


class TestUpperHemisphereIlluminanceLagarde2016(unittest.TestCase):
    """
    Define :func:`colour_hdri.calibration.absolute_luminance.\
upper_hemisphere_illuminance_Lagarde2016` definition unit tests methods.
    """

    def test_upper_hemisphere_illuminance_Lagarde2016(self):
        """
        Test :func:`colour_hdri.calibration.absolute_luminance.\
upper_hemisphere_illuminance_Lagarde2016` definition.
        """

        self.assertAlmostEqual(
            upper_hemisphere_illuminance_Lagarde2016(np.ones((16, 32, 3))),
            2.934469165342606,
            places=7,
        )

        self.assertAlmostEqual(
            upper_hemisphere_illuminance_Lagarde2016(
                np.ones((16, 32, 3)) * 10
            ),
            29.344691653426061,
            places=7,
        )

        self.assertAlmostEqual(
            upper_hemisphere_illuminance_Lagarde2016(
                np.ones((16, 32, 3)) * 0.1
            ),
            0.293446916534261,
            places=7,
        )


class TestUpperHemisphereIlluminanceWeightsLagarde2016(unittest.TestCase):
    """
    Define :func:`colour_hdri.calibration.absolute_luminance.\
upper_hemisphere_illuminance_weights_Lagarde2016` definition unit tests
    methods.
    """

    def test_upper_hemisphere_illuminance_weights_Lagarde2016(self):
        """
        Test :func:`colour_hdri.calibration.absolute_luminance.\
upper_hemisphere_illuminance_weights_Lagarde2016` definition.
        """

        weights = np.array(
            [
                [0.00000000],
                [1.98673676],
                [3.89213628],
                [5.63819129],
                [7.15341808],
                [8.37578310],
                [9.25524257],
                [9.75579132],
                [9.85693683],
                [9.55453819],
                [8.86097564],
                [7.80464369],
                [6.42878872],
                [4.78973839],
                [2.95459560],
                [0.99849132],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
                [0.00000000],
            ]
        )

        np.testing.assert_almost_equal(
            upper_hemisphere_illuminance_weights_Lagarde2016(32, 16),
            np.tile(weights, (1, 16)),
            decimal=7,
        )


class TestAbsoluteLuminanceCalibrationLagarde2016(unittest.TestCase):
    """
    Define :func:`colour_hdri.calibration.absolute_luminance.\
absolute_luminance_calibration_Lagarde2016` definition unit tests methods.
    """

    def test_absolute_luminance_calibration_Lagarde2016(self):
        """
        Test :func:`colour_hdri.calibration.absolute_luminance.\
absolute_luminance_calibration_Lagarde2016` definition.
        """

        # Unity Technologies. (2016). Treasure Island - white balanced.exr.
        # Retrieved August 30, 2016, from http://blog.selfshadow.com/\
        # publications/s2016-shading-course/unity/supplemental/\
        # Treasure Island - white balanced.exr

        reference_exr_file = read_image(
            str(
                os.path.join(
                    UNITY_001_DIRECTORY,
                    "Unity_Treasure_Island_White_Balanced.exr",
                )
            )
        )

        test_exr_file = read_image(
            str(
                os.path.join(
                    CALIBRATION_DIRECTORY,
                    "Unity_Treasure_Island_White_Balanced_Absolute.exr",
                )
            )
        )

        np.testing.assert_allclose(
            absolute_luminance_calibration_Lagarde2016(
                reference_exr_file, 51000
            ),
            test_exr_file,
            rtol=0.0000001,
            atol=0.0000001,
        )


if __name__ == "__main__":
    unittest.main()



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour_hdri.utilities.image` module."""

from __future__ import annotations

import numpy as np
import os
import unittest

from colour_hdri import TESTS_RESOURCES_DIRECTORY
from colour_hdri.utilities import filter_files
from colour_hdri.utilities import Image, ImageStack

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestImage",
    "TestImageStack",
]

FROBISHER_001_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "frobisher_001"
)


class TestImage(unittest.TestCase):
    """
    Define :class:`colour_hdri.utilities.image.Image` class unit tests
    methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._test_jpg_image = filter_files(FROBISHER_001_DIRECTORY, ("jpg",))[
            0
        ]

    def test_required_attributes(self):
        """Test the presence of required attributes."""

        required_attributes = ("path", "data", "metadata")

        for attribute in required_attributes:
            self.assertIn(attribute, dir(Image))

    def test_required_methods(self):
        """Test the presence of required methods."""

        required_methods = ("__init__", "read_data", "read_metadata")

        for method in required_methods:
            self.assertIn(method, dir(Image))

    def test_read_data(self):
        """Test :attr:`colour_hdri.utilities.image.Image.read_data` method."""

        image = Image(self._test_jpg_image)

        self.assertIsNone(image.data)
        self.assertTupleEqual(image.read_data().shape, (426, 640, 3))

    def test_read_metadata(self):
        """Test :attr:`colour_hdri.utilities.image.Image.end` method."""

        image = Image(self._test_jpg_image)

        self.assertEqual(image.metadata, None)
        np.testing.assert_array_equal(
            np.array(image.read_metadata()),
            np.array([8.0, 0.125, 100.0, np.nan, np.nan, np.nan]),
        )


class TestImageStack(unittest.TestCase):
    """
    Define :class:`colour_hdri.utilities.image.ImageStack` class unit tests
    methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._test_jpg_images = filter_files(FROBISHER_001_DIRECTORY, ("jpg",))

        self._image_stack = ImageStack().from_files(self._test_jpg_images)

    def test_required_methods(self):
        """Test the presence of required methods."""

        required_methods = (
            "__init__",
            "__getitem__",
            "__setitem__",
            "__delitem__",
            "__len__",
            "__getattr__",
            "__setattr__",
            "insert",
            "from_files",
        )

        for method in required_methods:
            self.assertIn(method, dir(ImageStack))

    def test__getitem__(self):
        """
        Test :attr:`colour_hdri.utilities.image.ImageStack.__getitem__`
        method.
        """

        for image in self._image_stack:
            self.assertIsInstance(image, Image)

    def test__setitem__(self):
        """
        Test :attr:`colour_hdri.utilities.image.ImageStack.__setitem__`
        method.
        """

        image_stack = ImageStack()
        image = Image(self._test_jpg_images[0])
        image.read_data()
        image.read_metadata()
        image_stack.insert(0, image)

        self.assertEqual(image_stack[0], image)

    def test__delitem__(self):
        """
        Test :attr:`colour_hdri.utilities.image.ImageStack.__delitem__`
        method.
        """

        image_stack = ImageStack().from_files(self._test_jpg_images)

        del image_stack[0]

        self.assertEqual(len(image_stack), 2)

    def test__len__(self):
        """Test :attr:`colour_hdri.utilities.image.ImageStack.__len__` method."""

        self.assertEqual(len(self._image_stack), 3)

    def test__getattr__(self):
        """
        Test :attr:`colour_hdri.utilities.image.ImageStack.__getattr__`
        method.
        """

        self.assertTupleEqual(self._image_stack.data.shape, (426, 640, 3, 3))

        np.testing.assert_almost_equal(
            self._image_stack.f_number, np.array([8, 8, 8]), decimal=7
        )

        self.assertEqual(self._image_stack[0].metadata.f_number, 8)

        np.testing.assert_almost_equal(
            self._image_stack.exposure_time, np.array([0.125, 1, 8]), decimal=7
        )

        self.assertEqual(self._image_stack[0].metadata.exposure_time, 0.125)

        np.testing.assert_array_equal(
            self._image_stack.black_level, np.array([np.nan, np.nan, np.nan])
        )

        self.assertEqual(self._image_stack[0].metadata.black_level, None)

    def test__setattr__(self):
        """
        Test :attr:`colour_hdri.utilities.image.ImageStack.__getattr__`
        method.
        """

        image_stack = ImageStack().from_files(self._test_jpg_images)

        self.assertTupleEqual(image_stack.data.shape, (426, 640, 3, 3))

        image_stack.data = np.random.random((20, 10, 3, 3))

        self.assertTupleEqual(image_stack.data.shape, (20, 10, 3, 3))

        np.testing.assert_almost_equal(
            image_stack.f_number, np.array([8, 8, 8]), decimal=7
        )

        image_stack.f_number = np.array([1, 2, 3])

        np.testing.assert_almost_equal(
            image_stack.f_number, np.array([1, 2, 3]), decimal=7
        )

        self.assertEqual(image_stack[0].metadata.f_number, 1)

        np.testing.assert_array_equal(
            image_stack.black_level, np.array([np.nan, np.nan, np.nan])
        )

        image_stack.black_level = np.array([2048, 2048, 2048])

        np.testing.assert_almost_equal(
            image_stack.black_level, np.array([2048, 2048, 2048]), decimal=7
        )

        self.assertEqual(image_stack[0].metadata.black_level, 2048)

    def test_from_files(self):
        """
        Test :attr:`colour_hdri.utilities.image.ImageStack.test_from_files`
        method.
        """

        image_stack = ImageStack().from_files(reversed(self._test_jpg_images))
        self.assertListEqual(list(image_stack.path), self._test_jpg_images)


if __name__ == "__main__":
    unittest.main()



# !/usr/bin/env python
"""
Defines the unit tests for the :mod:`colour_hdri.sampling.grossberg2003`
module.
"""

from __future__ import annotations

import numpy as np
import os
import unittest

from colour.hints import List

from colour_hdri import TESTS_RESOURCES_DIRECTORY
from colour_hdri.sampling import samples_Grossberg2003
from colour_hdri.utilities import ImageStack, filter_files

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "FROBISHER_001_DIRECTORY",
    "SAMPLING_DIRECTORY",
    "JPG_IMAGES",
    "TestSamplesGrossberg2003",
]

FROBISHER_001_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "frobisher_001"
)

SAMPLING_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "colour_hdri", "sampling"
)

JPG_IMAGES: List[str] = filter_files(FROBISHER_001_DIRECTORY, ("jpg",))


class TestSamplesGrossberg2003(unittest.TestCase):
    """
    Define :func:`colour_hdri.sampling.grossberg2003.\
samples_Grossberg2003` definition unit tests methods.
    """

    def test_samples_Grossberg2003(self):
        """
        Test :func:`colour_hdri.sampling.grossberg2003.\
samples_Grossberg2003` definition.
        """

        np.testing.assert_almost_equal(
            samples_Grossberg2003(ImageStack.from_files(JPG_IMAGES).data),
            np.load(
                os.path.join(
                    SAMPLING_DIRECTORY, "test_samples_Grossberg2003.npy"
                )
            ),
            decimal=7,
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['colour_hdri',
 'colour_hdri.calibration',
 'colour_hdri.calibration.tests',
 'colour_hdri.exposure',
 'colour_hdri.exposure.tests',
 'colour_hdri.generation',
 'colour_hdri.generation.tests',
 'colour_hdri.models',
 'colour_hdri.models.datasets',
 'colour_hdri.models.tests',
 'colour_hdri.plotting',
 'colour_hdri.process',
 'colour_hdri.process.tests',
 'colour_hdri.recovery',
 'colour_hdri.recovery.tests',
 'colour_hdri.sampling',
 'colour_hdri.sampling.tests',
 'colour_hdri.tonemapping',
 'colour_hdri.tonemapping.global_operators',
 'colour_hdri.tonemapping.global_operators.tests',
 'colour_hdri.utilities',
 'colour_hdri.utilities.tests']

package_data = \
{'': ['*'],
 'colour_hdri': ['examples/*', 'resources/colour-hdri-tests-datasets/*']}

install_requires = \
['colour-science>=0.4.0',
 'imageio>=2,<3',
 'numpy>=1.19,<2',
 'scipy>=1.5,<2',
 'typing-extensions>=4,<5']

extras_require = \
{'development': ['biblib-simple',
                 'black',
                 'coverage!=6.3',
                 'coveralls',
                 'flake8',
                 'flynt',
                 'invoke',
                 'jupyter',
                 'mypy',
                 'pre-commit',
                 'pydata-sphinx-theme',
                 'pydocstyle',
                 'pytest',
                 'pytest-cov',
                 'restructuredtext-lint',
                 'sphinx>=4,<5',
                 'sphinxcontrib-bibtex',
                 'toml',
                 'twine'],
 'optional': ['colour-demosaicing>=0.2.0'],
 'plotting': ['matplotlib>=3.2,!=3.5.0,!=3.5.1'],
 'read-the-docs': ['matplotlib>=3.2,!=3.5.0,!=3.5.1',
                   'pydata-sphinx-theme',
                   'sphinxcontrib-bibtex']}

setup_kwargs = {
    'name': 'colour-hdri',
    'version': '0.2.0',
    'description': 'HDRI / Radiance image processing algorithms for Python',
    'long_description': "Colour - HDRI\n=============\n\n.. start-badges\n\n|actions| |coveralls| |codacy| |version|\n\n.. |actions| image:: https://img.shields.io/github/workflow/status/colour-science/colour-hdri/Continuous%20Integration?label=actions&logo=github&style=flat-square\n    :target: https://github.com/colour-science/colour-hdri/actions\n    :alt: Develop Build Status\n.. |coveralls| image:: http://img.shields.io/coveralls/colour-science/colour-hdri/develop.svg?style=flat-square\n    :target: https://coveralls.io/r/colour-science/colour-hdri\n    :alt: Coverage Status\n.. |codacy| image:: https://img.shields.io/codacy/grade/290ad2c23b0749b99a1d548ca47d9062/develop.svg?style=flat-square\n    :target: https://www.codacy.com/app/colour-science/colour-hdri\n    :alt: Code Grade\n.. |version| image:: https://img.shields.io/pypi/v/colour-hdri.svg?style=flat-square\n    :target: https://pypi.org/project/colour-hdri\n    :alt: Package Version\n\n.. end-badges\n\nA `Python <https://www.python.org/>`__ package implementing various\nHDRI / Radiance image processing algorithms.\n\nIt is open source and freely available under the\n`New BSD License <https://opensource.org/licenses/BSD-3-Clause>`__ terms.\n\n..  image:: https://raw.githubusercontent.com/colour-science/colour-hdri/master/docs/_static/Radiance_001.png\n\n.. contents:: **Table of Contents**\n    :backlinks: none\n    :depth: 2\n\n.. sectnum::\n\nFeatures\n--------\n\nThe following features are available:\n\n- HDRI / Radiance Image Generation\n- Debevec (1997) Camera Response Function Computation\n- Grossberg (2003) Histogram Based Image Sampling\n- Variance Minimization Light Probe Sampling\n- Global Tonemapping Operators\n- Adobe DNG SDK Colour Processing\n- Absolute Luminance Calibration\n- Digital Still Camera (DSC) Exposure Model\n- Raw Processing Helpers\n\nExamples\n^^^^^^^^\n\nVarious usage examples are available from the\n`examples directory <https://github.com/colour-science/colour-hdri/tree/master/colour_hdri/examples>`__.\n\nUser Guide\n----------\n\nInstallation\n^^^^^^^^^^^^\n\nBecause of their size, the resources dependencies needed to run the various\nexamples and unit tests are not provided within the Pypi package. They are\nseparately available as\n`Git Submodules <https://git-scm.com/book/en/v2/Git-Tools-Submodules>`__\nwhen cloning the\n`repository <https://github.com/colour-science/colour-hdri>`__.\n\nPrimary Dependencies\n~~~~~~~~~~~~~~~~~~~~\n\n**Colour - HDRI** requires various dependencies in order to run:\n\n- `python >= 3.8, < 4 <https://www.python.org/download/releases/>`__\n- `colour-science <https://pypi.org/project/colour-science/>`__\n- `recordclass <https://pypi.org/project/recordclass/>`__\n\nOptional Features Dependencies\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n- `colour-demosaicing <https://pypi.org/project/colour-demosaicing/>`__\n- `Adobe DNG Converter <https://www.adobe.com/support/downloads/product.jsp?product=106&platform=Mac>`__\n- `dcraw <https://www.cybercom.net/~dcoffin/dcraw/>`__\n- `ExifTool <http://www.sno.phy.queensu.ca/~phil/exiftool/>`__\n- `rawpy <https://pypi.org/project/rawpy/>`__\n\nPypi\n~~~~\n\nOnce the dependencies are satisfied, **Colour - HDRI** can be installed from\nthe `Python Package Index <http://pypi.python.org/pypi/colour-hdri>`__ by\nissuing this command in a shell::\n\n    pip install --user colour-hdri\n\nThe optional features dependencies are installed as follows::\n\n    pip install --user 'colour-hdri[optional]'\n\nThe figures plotting dependencies are installed as follows::\n\n    pip install --user 'colour-hdri[plotting]'\n\nThe overall development dependencies are installed as follows::\n\n    pip install --user 'colour-hdri[development]'\n\nContributing\n^^^^^^^^^^^^\n\nIf you would like to contribute to `Colour - HDRI <https://github.com/colour-science/colour-hdri>`__,\nplease refer to the following `Contributing <https://www.colour-science.org/contributing/>`__\nguide for `Colour <https://github.com/colour-science/colour>`__.\n\nBibliography\n^^^^^^^^^^^^\n\nThe bibliography is available in the repository in\n`BibTeX <https://github.com/colour-science/colour-hdri/blob/develop/BIBLIOGRAPHY.bib>`__\nformat.\n\nAPI Reference\n-------------\n\nThe main technical reference for `Colour - HDRI <https://github.com/colour-science/colour-hdri>`__\nis the `API Reference <https://colour-hdri.readthedocs.io/en/latest/reference.html>`__.\n\nSee Also\n--------\n\nPublications\n^^^^^^^^^^^^\n\n- `Advanced High Dynamic Range Imaging: Theory and Practice <https://dl.acm.org/doi/book/10.5555/1996408>`__ by Banterle, F. et al.\n\n*Advanced High Dynamic Range Imaging: Theory and Practice* was used as a\nreference for some of the algorithms of **Colour - HDRI**.\n\nSoftware\n^^^^^^^^\n\n**C/C++**\n\n- `OpenCV <https://opencv.org/>`__ by Bradski, G.\n- `Piccante <https://github.com/cnr-isti-vclab/piccante>`__ by Banterle, F. and Benedetti, L.,\n\n*Piccante* was used to verify the Grossberg (2003) Histogram Based Image Sampling.\n\n**Matlab**\n\n- `HDR Toolbox <https://github.com/banterle/HDR_Toolbox>`__ by Banterle, F. et al.\n\nCode of Conduct\n---------------\n\nThe *Code of Conduct*, adapted from the `Contributor Covenant 1.4 <https://www.contributor-covenant.org/version/1/4/code-of-conduct.html>`__,\nis available on the `Code of Conduct <https://www.colour-science.org/code-of-conduct/>`__ page.\n\nContact & Social\n----------------\n\nThe *Colour Developers* can be reached via different means:\n\n- `Email <mailto:colour-developers@colour-science.org>`__\n- `Discourse <https://colour-science.discourse.group/>`__\n- `Facebook <https://www.facebook.com/python.colour.science>`__\n- `Github Discussions <https://github.com/colour-science/colour-hdri/discussions>`__\n- `Gitter <https://gitter.im/colour-science/colour>`__\n- `Twitter <https://twitter.com/colour_science>`__\n\nAbout\n-----\n\n| **Colour - HDRI** by Colour Developers\n| Copyright 2015 Colour Developers – `colour-developers@colour-science.org <colour-developers@colour-science.org>`__\n| This software is released under terms of New BSD License: https://opensource.org/licenses/BSD-3-Clause\n| `https://github.com/colour-science/colour-hdri <https://github.com/colour-science/colour-hdri>`__\n",
    'author': 'Colour Developers',
    'author_email': 'colour-developers@colour-science.org',
    'maintainer': 'Colour Developers',
    'maintainer_email': 'colour-developers@colour-science.org',
    'url': 'https://www.colour-science.org/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from .weighting_functions import (
    normal_distribution_function,
    hat_function,
    weighting_function_Debevec1997,
)
from .radiance import image_stack_to_radiance_image

__all__ = []
__all__ += [
    "normal_distribution_function",
    "hat_function",
    "weighting_function_Debevec1997",
]
__all__ += [
    "image_stack_to_radiance_image",
]



from .radiance import plot_radiance_image_strip
from .tonemapping import plot_tonemapping_operator_image

__all__ = []
__all__ += [
    "plot_radiance_image_strip",
]
__all__ += [
    "plot_tonemapping_operator_image",
]



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour_hdri.process.adobe_dng` module."""

from __future__ import annotations

import numpy as np
import os
import platform
import shutil
import tempfile
import unittest
import zipfile

from colour import read_image
from colour.hints import List

from colour_hdri import TESTS_RESOURCES_DIRECTORY
from colour_hdri.process import (
    convert_raw_files_to_dng_files,
    convert_dng_files_to_intermediate_files,
    read_dng_files_exif_tags,
)
from colour_hdri.utilities import filter_files

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "FROBISHER_001_DIRECTORY",
    "PROCESS_DIRECTORY",
    "RAW_IMAGES",
    "TestConvertRawFilesToDngFiles",
    "TestConvertDngFilesToIntermediateFiles",
]

FROBISHER_001_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "frobisher_001"
)

PROCESS_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "colour_hdri", "process"
)

RAW_IMAGES: List[str] = filter_files(FROBISHER_001_DIRECTORY, ("CR2",))


class TestConvertRawFilesToDngFiles(unittest.TestCase):
    """
    Define :func:`colour_hdri.process.adobe_dng.\
convert_raw_files_to_dng_files` definition unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._temporary_directory = tempfile.mkdtemp()

    def tearDown(self):
        """After tests actions."""

        shutil.rmtree(self._temporary_directory)

    def test_convert_raw_files_to_dng_files(self):
        """
        Test :func:`colour_hdri.process.adobe_dng.\
convert_raw_files_to_dng_files` definition.
        """

        if platform.system() not in ("Windows", "Microsoft"):
            # *Adobe DNG Converter* is not available on *Linux*, and is not
            # reproducible on *macOS* thus we skip this unit test.
            return

        reference_dng_files = sorted(filter_files(PROCESS_DIRECTORY, ("dng",)))
        test_dng_files = sorted(
            convert_raw_files_to_dng_files(
                RAW_IMAGES, self._temporary_directory
            )
        )

        for test_dng_file, reference_dng_file in zip(
            test_dng_files, reference_dng_files
        ):
            np.testing.assert_almost_equal(
                read_image(str(test_dng_file)),
                read_image(str(reference_dng_file)),
                decimal=7,
            )


class TestConvertDngFilesToIntermediateFiles(unittest.TestCase):
    """
    Define :func:`colour_hdri.process.adobe_dng.\
convert_dng_files_to_intermediate_files` definition unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._temporary_directory = tempfile.mkdtemp()

    def tearDown(self):
        """After tests actions."""

        shutil.rmtree(self._temporary_directory)

    def test_convert_dng_files_to_intermediate_files(self):
        """
        Test :func:`colour_hdri.process.adobe_dng.\
convert_dng_files_to_intermediate_files` definition.
        """

        reference_dng_files = sorted(filter_files(PROCESS_DIRECTORY, ("dng",)))
        tests_dng_files = [
            os.path.join(
                self._temporary_directory, os.path.basename(reference_dng_file)
            )
            for reference_dng_file in reference_dng_files
        ]
        for reference_dng_file, tests_dng_file in zip(
            reference_dng_files, tests_dng_files
        ):
            shutil.copyfile(reference_dng_file, tests_dng_file)

        reference_zip_files = sorted(filter_files(PROCESS_DIRECTORY, ("zip",)))

        for reference_zip_file in reference_zip_files:
            with zipfile.ZipFile(reference_zip_file) as zip_file:
                tiff_file_name = os.path.basename(reference_zip_file).replace(
                    ".zip", ""
                )
                with open(
                    os.path.join(self._temporary_directory, tiff_file_name),
                    "wb",
                ) as reference_tiff_file:
                    reference_tiff_file.write(zip_file.read(tiff_file_name))

        reference_tiff_files = sorted(
            filter_files(self._temporary_directory, ("tiff",))
        )

        # for reference_tiff_file in reference_tiff_files:
        #     os.chdir(os.path.dirname(reference_tiff_file))
        #     with zipfile.ZipFile(
        #             '{0}.zip'.format(reference_tiff_file),
        #             mode='w') as zip_file:
        #         zip_file.write(
        #             os.path.basename(reference_tiff_file),
        #             compress_type=zipfile.ZIP_DEFLATED)
        #         os.remove(reference_tiff_file)

        test_tiff_files = sorted(
            convert_dng_files_to_intermediate_files(
                tests_dng_files, self._temporary_directory
            )
        )

        for test_tiff_file, reference_tiff_file in zip(
            test_tiff_files, reference_tiff_files
        ):
            np.testing.assert_almost_equal(
                read_image(str(test_tiff_file)),
                read_image(str(reference_tiff_file)),
                decimal=7,
            )


class TestReadDngFilesExifTags(unittest.TestCase):
    """
    Define :func:`colour_hdri.process.adobe_dng.\
read_dng_files_exif_tags` definition unit tests methods.
    """

    def test_read_dng_files_exif_tags(self):
        """
        Test :func:`colour_hdri.process.adobe_dng.\
read_dng_files_exif_tags` definition.
        """

        reference_dng_files = sorted(filter_files(PROCESS_DIRECTORY, ("dng",)))
        exif_tags = read_dng_files_exif_tags(reference_dng_files)
        self.assertEqual(len(exif_tags), 3)
        self.assertIn("EXIF", exif_tags[0])
        self.assertIn("Make", exif_tags[0]["EXIF"])

        self.assertAlmostEqual(
            exif_tags[0]["EXIF"]["Exposure Time"], 0.12500000, places=7
        )

        np.testing.assert_array_equal(
            exif_tags[0]["EXIF"]["Reduction Matrix 1"], np.identity(3)
        )


if __name__ == "__main__":
    unittest.main()



from .dng import (
    CCS_ILLUMINANT_ADOBEDNG,
    CCT_ILLUMINANTS_ADOBEDNG,
    LIGHT_SOURCE_TAG_TO_DNG_ILLUMINANTS,
)

__all__ = [
    "CCS_ILLUMINANT_ADOBEDNG",
    "CCT_ILLUMINANTS_ADOBEDNG",
    "LIGHT_SOURCE_TAG_TO_DNG_ILLUMINANTS",
]



# !/usr/bin/env python
"""
Defines the unit tests for the :mod:`colour_hdri.calibration.debevec1997`
module.
"""

from __future__ import annotations

import numpy as np
import os
import unittest

from colour.hints import List

from colour_hdri import TESTS_RESOURCES_DIRECTORY
from colour_hdri.calibration import (
    g_solve,
    camera_response_functions_Debevec1997,
)
from colour_hdri.exposure import average_luminance
from colour_hdri.sampling import samples_Grossberg2003
from colour_hdri.utilities import ImageStack, filter_files

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "FROBISHER_001_DIRECTORY",
    "CALIBRATION_DIRECTORY",
    "JPG_IMAGES",
    "TestGSolve",
    "TestCameraResponseFunctionsDebevec1997",
]

FROBISHER_001_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "frobisher_001"
)

CALIBRATION_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "colour_hdri", "calibration"
)

JPG_IMAGES: List[str] = filter_files(FROBISHER_001_DIRECTORY, ("jpg",))


class TestGSolve(unittest.TestCase):
    """
    Define :func:`colour_hdri.calibration.debevec1997.g_solve` definition
    unit tests methods.
    """

    def test_g_solve(self):
        """Test :func:`colour_hdri.calibration.debevec1997.g_solve` definition."""

        image_stack = ImageStack.from_files(JPG_IMAGES)
        L_l = np.log(
            1
            / average_luminance(
                image_stack.f_number,
                image_stack.exposure_time,
                image_stack.iso,
            )
        )
        samples = samples_Grossberg2003(image_stack.data)

        for i in range(3):
            g, lE = g_solve(samples[..., i], L_l)

            # Lower precision for unit tests under *Github Actions*.
            np.testing.assert_allclose(
                g,
                np.load(
                    os.path.join(
                        CALIBRATION_DIRECTORY, f"test_g_solve_g_{i}.npy"
                    )
                ),
                rtol=0.001,
                atol=0.001,
            )

            # Lower precision for unit tests under *Github Actions*.
            np.testing.assert_allclose(
                lE,
                np.load(
                    os.path.join(
                        CALIBRATION_DIRECTORY, f"test_g_solve_lE_{i}.npy"
                    )
                ),
                rtol=0.001,
                atol=0.001,
            )


class TestCameraResponseFunctionsDebevec1997(unittest.TestCase):
    """
    Define :func:`colour_hdri.calibration.debevec1997.\
camera_response_functions_Debevec1997` definition unit tests methods.
    """

    def test_camera_response_function_Debevec1997(self):
        """
        Test :func:`colour_hdri.calibration.debevec1997.\
camera_response_functions_Debevec1997` definition.
        """

        # Lower precision for unit tests under *Github Actions*.
        np.testing.assert_allclose(
            camera_response_functions_Debevec1997(
                ImageStack.from_files(JPG_IMAGES)
            ),
            np.load(
                os.path.join(
                    CALIBRATION_DIRECTORY,
                    "test_camera_response_function_Debevec1997_crfs.npy",
                )
            ),
            rtol=0.00001,
            atol=0.00001,
        )


if __name__ == "__main__":
    unittest.main()
